#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.21 20:00:00                  #
# ================================================== #

from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_AGENT_LLAMA,
    MODE_EXPERT,
    MODE_LANGCHAIN,
    MODE_LLAMA_INDEX,
)
class Launcher:
    def __init__(self, window=None):
        """
        Launcher controller

        :param window: Window instance
        """
        self.window = window
        self.no_api_key_allowed = [
            MODE_LANGCHAIN,
            MODE_LLAMA_INDEX,
            MODE_AGENT,
            MODE_AGENT_LLAMA,
            MODE_EXPERT,
        ]

    def post_setup(self):
        """Post setup launcher"""
        # show welcome API KEY dialog (disable for langchain mode)
        if not self.window.core.config.get('mode') in self.no_api_key_allowed and \
                (self.window.core.config.get('api_key') is None or self.window.core.config.get('api_key') == ''):

            if not self.window.core.config.get('api_key.monit.displayed', False):
                self.show_api_monit()
                self.window.core.config.set('api_key.monit.displayed', True)

        # check for updates
        if self.window.core.config.get('updater.check.launch'):
            self.window.core.updater.check()

    def show_api_monit(self):
        """Show empty API KEY monit"""
        self.window.ui.dialogs.open('info.start')

    def check_updates(self):
        """Check for updates"""
        self.window.core.updater.check(True)

    def toggle_update_check(self, value):
        """Toggle update check on startup"""
        self.window.core.config.set('updater.check.launch', value)
        self.window.core.config.save()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 19:00:00                  #
# ================================================== #

from pygpt_net.core.events import (
    ControlEvent,
    AppEvent,
    Event,
    KernelEvent,
    RenderEvent,
    BaseEvent
)


class EventsDebug:
    def __init__(self, window=None):
        """
        Events debug

        :param window: Window instance
        """
        self.window = window
        self.id = 'events'

    def update(self):
        """Update debug window."""
        self.window.core.debug.begin(self.id)
        self.window.core.debug.add(self.id, 'App Events:', str(self.extract_events(AppEvent)))
        self.window.core.debug.add(self.id, 'Control Events:', str(self.extract_events(ControlEvent)))
        self.window.core.debug.add(self.id, 'Kernel Events:', str(self.extract_events(KernelEvent)))
        self.window.core.debug.add(self.id, 'Render Events:', str(self.extract_events(RenderEvent)))
        self.window.core.debug.add(self.id, 'Plugin Events:', str(self.extract_events(Event)))
        self.window.core.debug.add(self.id, '----', '')
        self.window.core.debug.add(self.id, 'Voice Cmds (all):', str(self.window.core.access.voice.commands))
        self.window.core.debug.add(self.id, 'Voice Cmds (allowed):', str(self.window.core.access.voice.get_commands()))
        self.window.core.debug.end(self.id)

    def extract_events(self, events: BaseEvent) -> dict:
        """
        Extract events

        :param events: Events class
        """
        result = {}
        for property, value in vars(events).items():
            if (isinstance(value, str)
                    or isinstance(value, int)
                    or isinstance(value, float)
                    or isinstance(value, bool)):
                result[property] = value
        return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.14 18:00:00                  #
# ================================================== #

import os
from pathlib import Path

from PySide6.QtWidgets import QApplication

from pygpt_net.utils import trans


class Workdir:
    def __init__(self, window=None):
        """
        Workdir migration controller

        :param window: Window instance
        """
        self.window = window
        self.is_dialog = False
        self.busy = False

    def change(self):
        """Change working directory (open dialog)"""
        if self.is_dialog:
            self.window.ui.dialogs.close('workdir.change')
            self.is_dialog = False
        else:
            self.window.ui.dialogs.workdir.prepare()
            self.window.ui.dialogs.workdir.set_path(self.window.core.config.get_user_path())
            self.window.ui.dialogs.workdir.hide_status()
            self.window.ui.dialogs.open(
                'workdir.change',
                width=600,
                height=180,
            )
            self.is_dialog = True

    def update(self, path: str, force: bool = False):
        """
        Switch working directory to the existing one

        :param path: existing working directory
        :param force: force update (confirm)
        """
        print("\n====================")
        print("Changing workdir to: ", path)
        print("====================\n")
        current_path = self.window.core.config.get_user_path()
        default_path = self.window.core.config.get_base_workdir()
        if force:
            self.window.ui.dialogs.workdir.show_status(trans("dialog.workdir.result.wait"))
            QApplication.processEvents()

        lock_file = os.path.join(default_path, 'path.cfg')  # put "path.cfg"
        lock_path = path.replace(str(Path.home()), "%HOME%")
        if path == default_path:
            lock_path = ""  # set empty if default dir
        with open(lock_file, 'w', encoding='utf-8') as f:
            f.write(lock_path)  # new path

        # update path in current profile
        self.window.core.config.profile.update_current_workdir(path)

        # reload config
        self.window.core.config.set_workdir(path, reload=True)
        self.window.core.config.set('license.accepted', True)  # accept license to prevent show dialog again

        # reload components
        if force:
            try:
                self.window.controller.reload()
                self.window.ui.dialogs.workdir.show_status(trans("dialog.workdir.result.success").format(path=path))
                self.window.ui.dialogs.alert(trans("dialog.workdir.result.success").format(path=path))
            except Exception as e:
                self.window.core.debug.log(e)
                self.window.ui.dialogs.alert(e)
                print("Error reloading components: ", e)
                self.restore(current_path)
                self.window.controller.reloading = False  # unlock
        else:
            self.window.controller.reload()

    def migrate(self, path: str, force: bool = False):
        """
        Migrate working directory

        :param path: new working directory
        :param force: force migration (confirm)
        """
        if self.busy:
            self.window.ui.dialogs.alert("Workdir migration in progress...")
            return

        # check if not the same directory
        current = self.window.core.config.get_user_path()
        if current == path:
            self.window.ui.dialogs.alert(trans("dialog.workdir.result.same_directory"))
            self.busy = False
            self.window.ui.dialogs.workdir.hide_status()
            return

        # check if path is not already a workdir
        if self.window.core.filesystem.is_workdir_in_path(path):
            self.window.ui.dialogs.confirm(
                type='workdir.update',
                id=path,
                msg=trans("dialog.workdir.update.confirm").format(path=path)
            )
            return

        # check if path is empty
        if not self.window.core.filesystem.is_directory_empty(path):
            self.window.ui.dialogs.alert(trans("dialog.workdir.change.empty.alert"))
            return

        # confirm move
        if not force:
            self.window.ui.dialogs.confirm(
                type='workdir.change',
                id=path,
                msg=trans("dialog.workdir.change.confirm").format(path=path)
            )
            return

        self.window.ui.dialogs.workdir.hide_status()
        self.busy = True
        print("Migrating workdir from: ", current, " to: ", path)

        # check if path exists
        if not os.path.exists(path) or not os.path.isdir(path):
            self.window.ui.dialogs.alert(trans("dialog.workdir.result.directory_not_exists"))
            self.window.ui.dialogs.workdir.hide_status()
            self.busy = False
            return

        # check free space
        space_required = self.window.core.filesystem.get_directory_size(current, human_readable=False)
        space_free = self.window.core.filesystem.get_free_disk_space(path, human_readable=False)
        if space_required > space_free:
            self.window.ui.dialogs.alert(trans("dialog.workdir.result.no_free_space").format(
                required=self.window.core.filesystem.get_directory_size(current),
                free=self.window.core.filesystem.get_free_disk_space(path),
            ))
            self.busy = False
            self.window.ui.dialogs.workdir.hide_status()
            return

        self.window.ui.dialogs.workdir.show_status(trans("dialog.workdir.result.wait"))
        QApplication.processEvents()

        # copy workdir
        try:
            result = self.window.core.filesystem.copy_workdir(current, path)
        except Exception as e:
            self.window.core.debug.log(e)
            self.window.ui.dialogs.alert(e)
            print("Error migrating workdir: ", e)
            result = False

        if result:
            try:
                # remove old workdir
                self.window.core.debug.info("Clearing old workdir: {}".format(current))
                try:
                    # allow errors here
                    self.window.core.filesystem.clear_workdir(current)
                except Exception as e:
                    self.window.core.debug.log(e)
                    print("Error clearing old workdir: ", e)
                self.update(path)  # update workdir to new path
                self.window.ui.dialogs.workdir.show_status(trans("dialog.workdir.result.success").format(path=path))
                self.window.ui.dialogs.alert(trans("dialog.workdir.result.success").format(path=path))
            except Exception as e:
                self.window.core.debug.log(e)
                self.window.ui.dialogs.alert(e)
                print("Error migrating workdir: ", e)
                self.restore(current)  # restore current if failed
                self.window.controller.reloading = False  # unlock
        else:
            self.window.ui.dialogs.workdir.show_status(trans("dialog.workdir.result.failed"))
            self.window.ui.dialogs.alert(trans("dialog.workdir.result.failed"))
            self.restore(current)  # restore current if failed
            self.window.controller.reloading = False  # unlock

        self.busy = False
        self.window.core.debug.info("Finished migrating workdir from: {} to: {}".format(current, path))

    def restore(self, current: str):
        """
        Restore default working directory

        :param current: current working directory
        """
        print("Reverting workdir to: ", current)
        self.window.core.config.set_workdir(current, reload=True)
        default_path = self.window.core.config.get_base_workdir()
        lock_file = os.path.join(default_path, 'path.cfg')
        lock_path = current.replace(str(Path.home()), "%HOME%")
        if current == default_path:
            lock_path = ""  # set empty if default dir
        with open(lock_file, 'w', encoding='utf-8') as f:
            f.write(lock_path)  # new path
        self.window.ui.dialogs.workdir.set_path(current)
        self.window.ui.dialogs.workdir.show_status("Failed. Reverted to current workdir: {}".format(current))
        self.window.controller.reload()
        self.window.core.config.profile.update_current_workdir(current)  # update profile



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.17 01:00:00                  #
# ================================================== #

import json
import ast
from datetime import datetime

from PySide6.QtCore import Qt, QAbstractTableModel, QModelIndex
from PySide6.QtGui import QCursor
from PySide6.QtWidgets import QAbstractItemView, QApplication, QMenu, \
    QSizePolicy, QTableView, QHeaderView

class DatabaseTableModel(QAbstractTableModel):
    def __init__(self, data, headers, timestamp_columns=None, convert_timestamps=True):
        super().__init__()
        self._data = data
        self._headers = headers
        self._timestamp_columns = timestamp_columns or []
        self._convert_timestamps = convert_timestamps

    def data(self, index, role):
        if role == Qt.DisplayRole:
            value = self._data[index.row()][index.column()]
            if self._headers[index.column()] in self._timestamp_columns and self._convert_timestamps:
                try:
                    if value != "None" and value != 0:
                        value = datetime.fromtimestamp(value).strftime('%Y-%m-%d %H:%M:%S')
                    else:
                        value = ""
                except Exception as e:
                    pass
            return value
        return None

    def rowCount(self, index=QModelIndex()):
        return len(self._data)

    def columnCount(self, index=QModelIndex()):
        return len(self._headers)

    def headerData(self, section, orientation, role):
        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            return self._headers[section]
        return None

class DatabaseList(QTableView):
    NAME = range(2)  # list of columns

    def __init__(self, window=None):
        """
        Database table records list

        :param window: Window instance
        """
        super(DatabaseList, self).__init__(window)
        self.window = window
        self.browser = None
        self.viewer = None
        self.viewer_index = None
        self.viewer_current = None
        self.viewer_current_id = None
        self.viewer_current_field = None
        self.selection = None
        self.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.verticalHeader().setVisible(False)
        self.setWordWrap(True)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.clicked.connect(self.onItemClicked)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(
            lambda: self.create_context_menu(self)
        )

    def on_data_begin(self):
        """
        On data update begin
        """
        self.adjustColumns()

    def on_data_end(self):
        """
        On data update end
        """
        # update data viewer
        if self.viewer is None:
            return
        if self.viewer_index:
            new = self.model().data(self.viewer_index)
            if self.viewer_current != new:
                self.viewer.setPlainText(self.parse_view(new))
                self.viewer_current = new

    def parse_view(self, data):
        """
        Parse view, convert data to formatted JSON if needed

        :param data: Data
        """
        try:
            if data.startswith("{") or data.startswith("["):
                data = json.dumps(ast.literal_eval(data), indent=4)
        except Exception as e:
            pass
        return data

    def adjustColumns(self):
        """Adjust columns width"""
        pass

    def create_context_menu(self, parent):
        """
        Create context menu

        :param parent: Parent
        """
        def copy_to_clipboard():
            index = parent.currentIndex()
            if index.isValid():
                value = index.sibling(index.row(), 1).data(
                    Qt.DisplayRole)
                QApplication.clipboard().setText(value)

        def delete_row():
            index = parent.currentIndex()
            if index.isValid():
                id = index.sibling(index.row(), 0).data(
                    Qt.DisplayRole)

                # delete row from database
                # self.window.get_viewer().delete_row(id)
                self.window.ui.dialogs.database.viewer.delete_row(int(id))

        menu = QMenu()
        copy_action = menu.addAction("Copy value to clipboard")
        delete_action = menu.addAction("Delete row")
        copy_action.triggered.connect(copy_to_clipboard)
        delete_action.triggered.connect(delete_row)
        menu.exec_(QCursor.pos())

    def mousePressEvent(self, event):
        """
        Mouse press event

        :param event: Event
        """
        index = self.indexAt(event.pos())
        if not index.isValid():
            return
        super(DatabaseList, self).mousePressEvent(event)

    def onItemClicked(self, index):
        """
        On item clicked

        :param index: Index
        """
        # update data viewer
        if self.viewer is None:
            return
        data = self.model().data(index, Qt.DisplayRole)
        id = index.sibling(index.row(), 0).data(
            Qt.DisplayRole)
        self.viewer.setPlainText(str(self.parse_view(data)))
        self.viewer_index = index
        self.viewer_current = data
        self.viewer_current_id = id
        self.viewer_current_field = self.model().headerData(index.column(), Qt.Horizontal, Qt.DisplayRole)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

from pathlib import Path


class Installer:

    def __init__(self, window=None):
        """
        Installer core

        :param window: Window instance
        """
        self.window = window

    def install(self):
        """Install db, config data and directories"""
        try:
            # create user config directory
            path = Path(self.window.core.config.path)
            path.mkdir(parents=True, exist_ok=True)

            # install config files and database
            self.window.core.config.install()

            # install models
            self.window.core.models.install()

            # install presets
            self.window.core.presets.install()

            # install indexes
            self.window.core.idx.install()

            # install history
            self.window.core.history.install()

            # install context
            self.window.core.ctx.install()

            # install notepad
            self.window.core.notepad.install()

            # install attachments
            self.window.core.attachments.install()

            # install assistants
            self.window.core.assistants.install()

            # install images
            self.window.core.image.install()

            # install filesystem
            self.window.core.filesystem.install()

            # install vision capture
            self.window.core.camera.install()

        except Exception as e:
            self.window.core.debug.log(e)
            print("Error installing config files:", e)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.15 01:00:00                  #
# ================================================== #

import json

from PySide6.QtCore import Slot

from pygpt_net.plugin.base.worker import BaseWorker, BaseSignals


class WorkerSignals(BaseSignals):
    pass  # add custom signals here


class Worker(BaseWorker):
    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = WorkerSignals()
        self.args = args
        self.kwargs = kwargs
        self.plugin = None
        self.websearch = None
        self.cmds = None
        self.ctx = None
        self.msg = None

    @Slot()
    def run(self):
        responses = []
        for item in self.cmds:
            if self.is_stopped():
                break
            response = None
            try:
                if item["cmd"] == "web_search":
                    response = self.cmd_web_urls(item)  # return URLs

                elif item["cmd"] == "web_url_open":
                    response = self.cmd_web_url_open(item)

                elif item["cmd"] == "web_url_raw":
                    response = self.cmd_web_url_raw(item)

                elif item["cmd"] == "web_urls":
                    response = self.cmd_web_urls(item)

                elif item["cmd"] == "web_index":
                    response = self.cmd_web_index(item)

                elif item["cmd"] == "web_index_query":
                    response = self.cmd_web_index_query(item)

                elif item["cmd"] == "web_extract_links":
                    response = self.cmd_web_extract_links(item)

                elif item["cmd"] == "web_extract_images":
                    response = self.cmd_web_extract_images(item)

                elif item["cmd"] == "web_request":
                    response = self.cmd_web_request(item)

                if response:
                    responses.append(response)

            except Exception as e:
                responses.append(
                    self.make_response(
                        item,
                        self.throw_error(e)
                    )
                )

        # send response
        if len(responses) > 0:
            self.reply_more(responses)

        if self.msg is not None:
            self.log(self.msg)
            self.status(self.msg)

    def cmd_web_search(self, item: dict) -> dict:
        """
        Web search command

        :param item: command item
        :return: response item
        """
        page = 1
        if self.has_param(item, "page"):
            page = int(self.get_param(item, "page"))
        prompt = None
        if self.has_param(item, "summarize_prompt"):
            prompt = self.get_param(item, "summarize_prompt")

        query = self.get_param(item, "query", "")
        content, total_found, current, url, img = self.websearch.make_query(
            query,
            page,
            prompt,
        )
        self.msg = "Web search finished: '{}'".format(query )
        result = {
            'query': query,
            'content': content,
            'url': url,
            'page': current,
            'total_found': total_found,
        }
        if url:
            self.ctx.urls_before.append(url)
        if img:
            result["thumb_img"] = img
            self.ctx.images_before.append(img)

        self.ctx.save_reply = True  # leave links in context
        return self.make_response(item, result)

    def cmd_web_url_open(self, item: dict) -> dict:
        """
        Open web URL command

        :param item: command item
        :return: response item
        """
        prompt = None
        if self.has_param(item, "summarize_prompt"):
            prompt = self.get_param(item, "summarize_prompt")
        url = self.get_param(item, "url", "")

        self.msg = "Opening Web URL: '{}'".format(url)
        content, url, img = self.websearch.open_url(
            url,
            prompt,
        )
        result = {
            'url': url,
            'content': content,
        }
        context = "From: " + url + ":\n--------------------------------\n" + str(content)
        if url:
            self.ctx.urls_before.append(url)
        if img:
            result["thumb_img"] = img
            self.ctx.images_before.append(img)

        extra = {
            "context": context,
        }
        return self.make_response(item, result, extra=extra)

    def cmd_web_url_raw(self, item: dict) -> dict:
        """
        Open web URL (raw) command

        :param item: command item
        :return: response item
        """
        url = self.get_param(item, "url", "")
        self.msg = "Opening Web URL: '{}'".format(url)
        content, url, img  = self.websearch.open_url_raw(
            url,
        )
        result = {
            'url': url,
            'content': content,
        }
        context = "From: " + url + ":\n--------------------------------\n" + str(content)
        if url:
            self.ctx.urls_before.append(url)
        if img:
            result["thumb_img"] = img
            self.ctx.images_before.append(img)

        extra = {
            "context": context,
        }
        return self.make_response(item, result, extra=extra)

    def cmd_web_urls(self, item: dict) -> dict:
        """
        Web search for URLs command

        :param item: command item
        :return: response item
        """
        page = 1
        num = 10
        if self.has_param(item, "page"):
            page = int(self.get_param(item, "page"))
        if self.has_param(item, "num_links"):
            num = int(self.get_param(item, "num_links"))
        if num < 1:
            num = 1
        if num > 10:
            num = 10
        offset = 1
        if page > 1:
            offset = (page - 1) * num + 1

        query = self.get_param(item, "query", "")
        urls = self.websearch.search(
            query,
            num,
            offset,
        )
        self.msg = "Web search finished: '{}'".format(query)
        result = {
            'query': query,
            'urls': json.dumps(urls),
            'page': page,
            'num': num,
            'offset': offset,
        }
        if urls:
            for url in urls:
                self.ctx.urls_before.append(url)

        return self.make_response(item, result)

    def cmd_web_index(self, item: dict) -> dict:
        """
        Index web URL command

        :param item: command item
        :return: response item
        """
        type = "webpage"  # default
        args = {}

        if self.has_param(item, "type"):
            type = self.get_param(item, "type")
        if self.has_param(item, "args"):
            args = self.get_param(item, "args")

        url = ""
        if self.has_param(item, "url"):
            url = self.get_param(item, "url")  # from default param
        if "url" in args:
            url = args["url"]  # override from args

        if not url:
            return self.make_response(item, "No URL provided")

        self.msg = "Indexing URL: '{}'".format(url)
        idx_name = self.plugin.get_option_value("idx")
        self.status("Please wait... indexing: {}...".format(url))

        # index URL via Llama-index
        num, errors = self.plugin.window.core.idx.index_urls(
            idx=idx_name,
            urls=[url],
            type=type,
            extra_args=args,
        )
        result = {
            'url': url,
            'num_indexed': num,
            'index': idx_name,
            'errors': errors,
        }
        if url and (url.startswith("http://") or url.startswith("https://")):
            self.ctx.urls_before.append(url)

        extra = {
            "url": url,
        }
        return self.make_response(item, result, extra=extra)

    def cmd_web_index_query(self, item: dict) -> dict:
        """
        Index web URL and query command

        :param item: command item
        :return: response item
        """
        type = "webpage"  # default
        args = {}
        query = None

        if self.has_param(item, "type"):
            type = self.get_param(item, "type")
        if self.has_param(item, "args"):
            args = self.get_param(item, "args")

        url = ""
        if self.has_param(item, "url"):
            url = self.get_param(item, "url")  # from default param
        if "url" in args:
            url = args["url"]  # override from args

        if not url:
            result = "No URL provided"
            return self.make_response(item, result)

        if self.has_param(item, "query"):
            query = self.get_param(item, "query")

        result = "No data"
        context = None
        if query is not None:
            # query file using temp index (created on the fly)
            self.log("Querying web: {}".format(url))

            # get tmp query model
            model = self.plugin.window.core.models.from_defaults()
            tmp_model = self.plugin.get_option_value("model_tmp_query")
            if self.plugin.window.core.models.has(tmp_model):
                model = self.plugin.window.core.models.get(tmp_model)

            answer = self.plugin.window.core.idx.chat.query_web(
                ctx=self.ctx,
                type=type,
                url=url,
                args=args,
                query=query,
                model=model,
            )
            self.log("Response from temporary in-memory index: {}".format(answer))
            if answer:
                from_str = type
                if url:
                    from_str += ", URL: " + url
                result = answer
                context = "From: " + from_str + ":\n--------------------------------\n" + answer

            # + auto-index to main index using Llama-index
            if self.plugin.get_option_value("auto_index"):
                self.msg = "Indexing URL: '{}'".format(url)
                idx_name = self.plugin.get_option_value("idx")
                # show status
                self.status("Please wait... indexing: {}...".format(url))
                # index URL via Llama-index
                num, errors = self.plugin.window.core.idx.index_urls(
                    idx=idx_name,
                    urls=[url],
                    type=type,
                    extra_args=args,
                )

        # add URL to context
        if url and (url.startswith("http://") or url.startswith("https://")):
            self.ctx.urls_before.append(url)

        extra = {
            "context": context,
        }
        return self.make_response(item, result, extra=extra)

    def cmd_web_extract_links(self, item: dict) -> dict:
        """
        Web extract links command

        :param item: command item
        :return: response item
        """
        url = ""
        if self.has_param(item, "url"):
            url = self.get_param(item, "url")
        if not url:
            return self.make_response(item, "No URL provided")
        links = self.plugin.window.core.web.helpers.get_links(url)
        result = {
            'links': links,
        }
        self.ctx.urls_before.append(url)
        return self.make_response(item, result)

    def cmd_web_extract_images(self, item: dict) -> dict:
        """
        Web extract images command

        :param item: command item
        :return: response item
        """
        download = False
        url = ""
        if self.has_param(item, "url"):
            url = self.get_param(item, "url")
        if self.has_param(item, "download"):
            download = bool(self.get_param(item, "download"))
        if not url:
            return self.make_response(item, "No URL provided")
        images = self.plugin.window.core.web.helpers.get_images(url)
        result = {
            'images': images,
        }
        if images and download:
            for img in images:
                try:
                    path = self.plugin.window.core.web.helpers.download_image(img)
                    if path:
                        self.ctx.images_before.append(path)
                except Exception as e:
                    print(e)
        self.ctx.urls_before.append(url)
        return self.make_response(item, result)

    def cmd_web_request(self, item: dict) -> dict:
        """
        Web request command

        :param item: command item
        :return: response item
        """
        url = ""
        method = "GET"
        data = None
        raw = None
        json = None
        headers = None
        params = None
        cookies = None
        files = None

        if self.has_param(item, "url"):
            url = self.get_param(item, "url")
        if not url:
            return self.make_response(item, "No URL provided")
        if self.has_param(item, "method"):
            method = self.get_param(item, "method")
        if self.has_param(item, "data_form"):
            data = self.get_param(item, "data_form")
        if self.has_param(item, "data"):
            raw = self.get_param(item, "data")
        if self.has_param(item, "data_json"):
            json = self.get_param(item, "data_json")
        if self.has_param(item, "headers"):
            headers = self.get_param(item, "headers")
        if self.has_param(item, "params"):
            params = self.get_param(item, "params")
        if self.has_param(item, "cookies"):
            cookies = self.get_param(item, "cookies")
        if self.has_param(item, "files"):
            files = self.get_param(item, "files")

        code, response = self.window.core.web.helpers.request(
            url=url,
            method=method,
            headers=headers,
            params=params,
            data=data if data else raw,
            json=json,
            cookies=cookies,
            files=files,
            timeout=self.plugin.get_option_value("timeout"),
            disable_ssl_verify=self.plugin.get_option_value("disable_ssl"),
            allow_redirects=True,
            stream=False,
            user_agent=self.plugin.get_option_value("user_agent"))
        result = {
            'url': url,
            'code': code,
            'response': response,
        }
        self.ctx.urls_before.append(url)
        return self.make_response(item, result)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

from typing import Dict

from packaging.version import Version

from pygpt_net.item.assistant import AssistantStoreItem


class BaseProvider:
    def __init__(self, window=None):
        self.window = window
        self.id = ""
        self.type = "assistant_store"

    def attach(self, window):
        self.window = window

    def install(self):
        pass

    def patch(self, version: Version) -> bool:
        pass

    def create(self, store: AssistantStoreItem) -> str:
        pass

    def load(self, id) -> AssistantStoreItem:
        pass

    def load_all(self) -> Dict[str, AssistantStoreItem]:
        pass

    def save(self, file:  AssistantStoreItem):
        pass

    def save_all(self, items: Dict[str, AssistantStoreItem]):
        pass

    def remove(self, id: str):
        pass

    def truncate(self):
        pass

    def dump(self, store: AssistantStoreItem) -> str:
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 00:00:00                  #
# ================================================== #

from typing import Any

from PySide6.QtCore import Qt
from PySide6.QtGui import QStandardItemModel

from pygpt_net.core.debug.agent import AgentDebug
from pygpt_net.core.debug.assistants import AssistantsDebug
from pygpt_net.core.debug.attachments import AttachmentsDebug
from pygpt_net.core.debug.config import ConfigDebug
from pygpt_net.core.debug.context import ContextDebug
from pygpt_net.core.debug.db import DatabaseDebug
from pygpt_net.core.debug.events import EventsDebug
from pygpt_net.core.debug.indexes import IndexesDebug
from pygpt_net.core.debug.kernel import KernelDebug
from pygpt_net.core.debug.models import ModelsDebug
from pygpt_net.core.debug.plugins import PluginsDebug
from pygpt_net.core.debug.presets import PresetsDebug
from pygpt_net.core.debug.tabs import TabsDebug
from pygpt_net.core.debug.ui import UIDebug


class Debug:
    DBG_KEY, DBG_VALUE = range(2)

    def __init__(self, window=None):
        """
        Debugger handler

        :param window: Window instance
        """
        self.window = window

        # setup workers
        self.workers = {}
        self.workers['agent'] = AgentDebug(self.window)
        self.workers['assistants'] = AssistantsDebug(self.window)
        self.workers['attachments'] = AttachmentsDebug(self.window)
        self.workers['config'] = ConfigDebug(self.window)
        self.workers['context'] = ContextDebug(self.window)
        self.workers['db'] = DatabaseDebug(self.window)
        self.workers['events'] = EventsDebug(self.window)
        self.workers['indexes'] = IndexesDebug(self.window)
        self.workers['kernel'] = KernelDebug(self.window)
        self.workers['models'] = ModelsDebug(self.window)
        self.workers['plugins'] = PluginsDebug(self.window)
        self.workers['presets'] = PresetsDebug(self.window)
        self.workers['tabs'] = TabsDebug(self.window)
        self.workers['ui'] = UIDebug(self.window)

        # prepare debug ids
        self.ids = self.workers.keys()
        self.models = {}
        self.initialized = {}
        self.active = {}
        self.idx = {}
        self.counters = {}

        # prepare debug workers data
        for id in self.ids:
            self.models[id] = self.create_model(self.window)
            self.initialized[id] = False
            self.active[id] = False
            self.idx[id] = 0

    def begin(self, id: str):
        """
        Begin debug data

        :param id: debug id
        """
        self.window.ui.debug[id].setModel(self.models[id])
        self.models[id].dataChanged.connect(self.window.ui.debug[id].on_data_begin)

        if id not in self.counters or self.counters[id] != self.models[id].rowCount():
            self.models[id].removeRows(0, self.models[id].rowCount())
            self.initialized[id] = False
        self.idx[id] = 0

    def end(self, id: str):
        """
        End debug data

        :param id: debug id
        """
        self.counters[id] = self.idx[id]
        self.initialized[id] = True
        self.window.ui.debug[id].on_data_end()

    def add(self, id: str, k: str, v: Any):
        """
        Append debug entry

        :param id: debug id
        :param k: key
        :param v: value
        """
        if self.initialized[id] is False:
            idx = self.models[id].rowCount()
            self.models[id].insertRow(idx)
            self.models[id].setData(self.models[id].index(idx, self.DBG_KEY), k)
            self.models[id].setData(self.models[id].index(idx, self.DBG_VALUE), v)
        else:
            for idx in range(0, self.models[id].rowCount()):
                if self.models[id].index(idx, self.DBG_KEY).data() == k:
                    self.models[id].setData(self.models[id].index(idx, self.DBG_VALUE), v)
                    self.idx[id] += 1
                    return
        self.idx[id] += 1

    def get_ids(self) -> list:
        """
        Get debug workers ids

        :return: list of ids
        """
        return list(self.ids)

    def get_workers(self) -> dict:
        """
        Get debug workers

        :return: dict of workers
        """
        return self.workers

    def update_worker(self, id: str):
        """
        Update debug worker

        :param id: debug id
        """
        if id in self.workers:
            self.workers[id].update()

    def is_active(self, id: str) -> bool:
        """
        Check if debug window is active

        :param id: debug id
        :return: True if active
        """
        if id not in self.active:
            return False
        return self.active[id]

    def show(self, id: str):
        """
        Activate debug window

        :param id: debug id
        """
        if id not in self.active:
            return
        self.active[id] = True
        self.window.ui.dialogs.open('debug.' + id, width=800, height=600)

    def hide(self, id: str):
        """
        Deactivate debug window

        :param id: debug id
        """
        if id not in self.active:
            return
        self.active[id] = False
        self.window.ui.dialogs.close('debug.' + id)

    def create_model(self, parent) -> QStandardItemModel:
        """
        Create list model

        :param parent: parent widget
        :return: model instance
        """
        model = QStandardItemModel(0, 2, parent)
        model.setHeaderData(self.DBG_KEY, Qt.Horizontal, "Key")
        model.setHeaderData(self.DBG_VALUE, Qt.Horizontal, "Value")
        return model



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.21 20:00:00                  #
# ================================================== #

from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_AGENT_LLAMA,
    MODE_CHAT,
    MODE_COMPLETION,
    MODE_LANGCHAIN,
    MODE_LLAMA_INDEX,
    MODE_VISION,
    MODE_RESEARCH,
)

class Vision:
    def __init__(self, window=None):
        """
        Chat vision controller

        :param window: Window instance
        """
        self.window = window
        self.is_enabled = False
        self.is_available = False
        self.allowed_modes = [
            MODE_CHAT,
            MODE_COMPLETION,
            MODE_LANGCHAIN,
            MODE_LLAMA_INDEX,
            MODE_AGENT,
            MODE_AGENT_LLAMA,
            MODE_RESEARCH,
        ]

    def setup(self):
        """Set up UI"""
        pass

    def show_inline(self):
        """Show inline vision checkbox"""
        self.window.ui.nodes['inline.vision'].setVisible(True)  # show vision checkbox

    def hide_inline(self):
        """Hide inline vision checkbox"""
        self.window.ui.nodes['inline.vision'].setVisible(False)  # hide vision checkbox

    def available(self):
        """Set vision content available"""
        self.is_available = True

    def unavailable(self):
        """Set vision content unavailable"""
        self.is_available = False

    def switch_to_vision(self):
        """Switch to vision mode"""
        mode = self.window.core.config.get('mode')
        model = self.window.core.config.get('model')
        model_data = self.window.core.models.get(model)
        if mode in [MODE_AGENT, MODE_AGENT_LLAMA]:
            return  # disallow change in agent modes
        if mode == MODE_CHAT and MODE_VISION in model_data.mode:
            return  # abort if vision is already allowed
        if mode == MODE_VISION:
            return
        # abort if vision is already enabled
        if not self.window.controller.plugins.is_enabled('openai_vision') \
                or (self.window.controller.plugins.is_enabled('openai_vision')
                    and mode not in self.allowed_modes):
            self.window.controller.mode.set(MODE_VISION)

    def allowed(self) -> bool:
        """
        Check if vision content is allowed

        :return: True if allowed
        """
        if self.window.controller.plugins.is_enabled('openai_vision') \
                or self.window.core.config.get('mode') in self.allowed_modes:
            return True
        return False




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Optional, List, Dict, Any

from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_AGENT_LLAMA,
)
from pygpt_net.core.events import KernelEvent
from pygpt_net.core.bridge import BridgeContext
from pygpt_net.core.ctx.reply import ReplyContext
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans


class Legacy:
    def __init__(self, window=None):
        """
        Agent flow controller

        :param window: Window instance
        """
        self.window = window
        self.iteration = 0  # legacy
        self.prev_output = None
        self.is_user = True
        self.stop = False
        self.finished = False
        self.allowed_cmds = [
            "goal_update",
        ]
        self.pause_status = ["pause", "failed", "wait"]
        self.prompt_goal_native = """
        STATUS UPDATE: You can use "goal_update" command to update status of the task. 
        ON GOAL FINISH: When you believe that the task has been completed 100% and all goals have been achieved, 
        run "goal_update" command with status = "finished".
        ON PAUSE, FAILED OR WAIT: If more data from user is needed to achieve the goal or task run must be paused 
        or task was failed or when the conversation falls into a loop, THEN STOP REASONING and include "goal_update" 
        command with one of these statuses: "pause", "failed" or "wait".
        """
        self.options = {
            "agent.iterations": {
                "type": "int",
                "slider": True,
                "label": "agent.iterations",
                "min": 0,
                "max": 100,
                "step": 1,
                "value": 3,
                "multiplier": 1,
            },
        }

    def setup(self):
        """Setup agent controller"""
        # register hooks
        self.window.ui.add_hook("update.global.agent.iterations", self.hook_update)
        self.reload()  # restore config

    def reload(self):
        """Reload agent toolbox options"""
        # auto-stop checkbox
        if self.window.core.config.get('agent.auto_stop'):
            self.window.ui.config['global']['agent.auto_stop'].setChecked(True)
        else:
            self.window.ui.config['global']['agent.auto_stop'].setChecked(False)

        # continuous mode checkbox
        if self.window.core.config.get('agent.continue.always'):
            self.window.ui.config['global']['agent.continue'].setChecked(True)
        else:
            self.window.ui.config['global']['agent.continue'].setChecked(False)

        # iterations slider
        self.window.controller.config.apply_value(
            parent_id="global",
            key="agent.iterations",
            option=self.options["agent.iterations"],
            value=self.window.core.config.get('agent.iterations'),
        )

    def update(self):
        """Update agent status"""
        iterations = "-"
        mode = self.window.core.config.get('mode')

        # get iterations from plugin or from agent mode
        if mode in [
            MODE_AGENT,
            MODE_AGENT_LLAMA,
        ]:
            iterations = int(self.window.core.config.get("agent.iterations"))
        elif self.is_inline():
            if self.window.controller.plugins.is_enabled("agent"):
                iterations = int(self.window.core.plugins.get_option("agent", "iterations"))
        if iterations == 0:
            iterations_str = "∞"  # infinity loop
        else:
            iterations_str = str(iterations)

        status = str(self.iteration) + " / " + iterations_str
        self.window.ui.nodes['status.agent'].setText(status)
        self.window.controller.agent.common.toggle_status()

    def get_functions(self) -> List[Dict[str, Any]]:
        """
        Append goal commands

        :return: goal commands
        """
        cmds = [
            {
                "cmd": "goal_update",
                "instruction": "Update goal status",
                "params": [
                    {
                        "name": "status",
                        "description": "status to update",
                        "required": True,
                        "type": "str",
                        "enum": {
                            "status": ["finished", "pause", "failed", "wait"],
                        }
                    }
                ]
            }
        ]
        return cmds

    def on_system_prompt(
            self,
            prompt: str,
            append_prompt: Optional[str] = "",
            auto_stop: bool = True,
    ) -> str:
        """
        Event: On prepare system prompt

        :param prompt: prompt
        :param append_prompt: extra prompt (instruction)
        :param auto_stop: auto stop
        :return: updated prompt
        """
        stop_cmd = ""
        if auto_stop:
            if not self.window.core.command.is_native_enabled():
                stop_cmd = "\n\n" + self.window.core.prompt.get("agent.goal")  # use ### commands
            else:
                stop_cmd = "\n\n" + self.prompt_goal_native  # use API native functions
        if append_prompt is not None and append_prompt.strip() != "":
            append_prompt = "\n" + append_prompt
        prompt += str(append_prompt) + stop_cmd
        return prompt

    def on_input_before(self, prompt: str) -> str:
        """
        Event: On user input before

        :param prompt: prompt
        :return: updated prompt
        """
        if not self.is_user:
            return prompt

        return "user: " + prompt  # add user prefix

    def on_user_send(self, text: str):
        """
        Event: On user send text

        :param text: text
        """
        self.iteration = 0
        self.prev_output = None
        self.is_user = True
        self.finished = False  # reset finished flag
        if self.stop:
            self.stop = False
        self.window.controller.agent.legacy.update()  # update status

    def on_ctx_end(
            self,
            ctx: CtxItem,
            iterations: int = 0,
    ):
        """
        Event: On context end

        :param ctx: CtxItem
        :param iterations: iterations
        """
        if self.stop:
            self.stop = False
            self.iteration = 0
            self.prev_output = None
            return

        # if ctx has commands to execute then abort sending reply (command response will be sent)
        reply_abort = False
        if len(ctx.cmds) > 0:
            reply_abort = True
            # exception for expert_call command
            if len(ctx.cmds) == 1 and ctx.cmds[0]["cmd"] == "expert_call":
                reply_abort = False
            # if commands execution is disabled, then don't abort (there is no command response)
            if not self.window.core.config.get("cmd"):
                reply_abort = False
        if reply_abort:
            return

        if iterations == 0 or self.iteration < int(iterations):
            self.iteration += 1
            self.window.controller.agent.legacy.update()  # update status
            if self.prev_output is not None and self.prev_output != "":

                # always abort if already waiting for reply from expert or command
                if not self.window.controller.kernel.stack.waiting():
                    # add to reply stack
                    reply = ReplyContext()
                    reply.type = ReplyContext.AGENT_CONTINUE
                    reply.ctx = ctx
                    reply.input = self.prev_output
                    # send to kernel
                    context = BridgeContext()
                    context.ctx = ctx
                    context.reply_context = reply
                    event = KernelEvent(KernelEvent.AGENT_CONTINUE, {
                        'context': context,
                        'extra': {},
                    })
                    self.window.dispatch(event)

        # internal call will not trigger async mode and will hide the message from previous iteration
        elif self.iteration >= int(iterations):
            self.on_stop(auto=True)
            if self.window.core.config.get("agent.goal.notify"):
                self.window.ui.tray.show_msg(
                    trans("notify.agent.stop.title"),
                    trans("notify.agent.stop.content"),
                )

    def on_ctx_before(
            self,
            ctx: CtxItem,
            reverse_roles: bool = False,
    ):
        """
        Event: Before ctx

        :param ctx: CtxItem
        :param reverse_roles: reverse roles
        """
        ctx.internal = True  # always force internal call
        self.is_user = False
        if self.iteration == 0:
            ctx.first = True

        # reverse roles in ctx
        if self.iteration > 0 \
                and self.iteration % 2 != 0 \
                and reverse_roles:
            tmp_input_name = ctx.input_name
            tmp_output_name = ctx.output_name
            ctx.input_name = tmp_output_name
            ctx.output_name = tmp_input_name

    def on_ctx_after(self, ctx: CtxItem):
        """
        Event: After ctx

        :param ctx: CtxItem
        """
        self.prev_output = self.window.core.prompt.get("agent.continue")  # continue if needed...
        if self.window.core.config.get('agent.continue.always'):
            self.prev_output = self.window.core.prompt.get("agent.continue.always")  # continue reasoning...
        if ctx.extra_ctx is not None and ctx.extra_ctx != "":
            self.prev_output = ctx.extra_ctx

    def on_cmd(
            self,
            ctx: CtxItem,
            cmds: List[Dict[str, Any]],
    ):
        """
        Event: On commands

        :param ctx: CtxItem
        :param cmds: commands dict
        """
        if self.window.core.config.get('agent.auto_stop'):
            self.cmd(ctx, cmds)

    def cmd(
            self,
            ctx: CtxItem,
            cmds: List[Dict[str, Any]],
    ):
        """
        Event: On command

        :param ctx: CtxItem
        :param cmds: commands dict
        """
        is_cmd = False
        my_commands = []
        for item in cmds:
            if item["cmd"] in self.allowed_cmds:
                my_commands.append(item)
                is_cmd = True

        if not is_cmd:
            return

        for item in my_commands:
            try:
                if item["cmd"] == "goal_update":
                    if item["params"]["status"] == "finished":
                        self.on_stop(auto=True)
                        self.window.update_status(trans('status.finished'))  # show info
                        self.finished = True
                        if self.window.core.config.get("agent.goal.notify"):
                            self.window.ui.tray.show_msg(
                                trans("notify.agent.goal.title"),
                                trans("notify.agent.goal.content"),
                            )
                    elif item["params"]["status"] in self.pause_status:
                        self.on_stop(auto=True)
                        self.window.update_status(trans('status.finished'))  # show info
                        self.finished = True
            except Exception as e:
                self.window.core.debug.error(e)
                return

    def is_inline(self) -> bool:
        """
        Is agent inline (plugin) enabled

        :return: True if enabled
        """
        return self.window.controller.plugins.is_type_enabled("agent")

    def enabled(self) -> bool:
        """
        Is agent enabled

        :return: True if enabled
        """
        return self.window.core.config.get('mode') == MODE_AGENT or self.is_inline()

    def add_run(self):
        """Increment agent iteration"""
        self.iteration += 1

    def on_stop(self, auto: bool = False):
        """
        Event: On force stop

        :param auto: auto
        """
        self.window.controller.kernel.stack.lock()
        self.window.controller.chat.common.unlock_input()
        self.iteration = 0
        self.prev_output = None
        self.stop = True
        self.finished = False  # reset finished flag

        # update index if auto-index enabled
        if auto:
            self.window.controller.idx.on_ctx_end(
                ctx=None,
                mode="agent",
            )

    def hook_update(self, key: str, value: Any, caller, *args, **kwargs):
        """
        Hook: on option update

        :param key: config key
        :param value: config value
        :param caller: caller name
        :param args: args
        :param kwargs: kwargs
        """
        if self.window.core.config.get(key) == value:
            return
        if key == 'agent.iterations':
            self.window.core.config.set(key, int(value))  # cast to int, if from text input
            self.window.core.config.save()
            self.update()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

import datetime
import uuid
from typing import Optional, Dict

from packaging.version import Version

from pygpt_net.core.tabs.tab import Tab
from pygpt_net.item.notepad import NotepadItem
from pygpt_net.provider.core.notepad.db_sqlite import DbSqliteProvider
from pygpt_net.utils import trans


class Notepad:
    def __init__(self, window=None):
        """
        Notepad core

        :param window: Window instance
        """
        self.window = window
        self.provider = DbSqliteProvider(window)
        self.items = {}
        self.locked = False

    def install(self):
        """Install provider data"""
        self.provider.install()

    def patch(self, app_version: Version) -> bool:
        """
        Patch provider data

        :param app_version: app version
        :return: True if data was patched
        """
        return self.provider.patch(app_version)

    def reset(self):
        """Reset provider data"""
        self.items = {}

    def get_by_id(self, idx: int) -> Optional[NotepadItem]:
        """
        Get notepad by idx

        :param idx: notepad idx
        :return: notepad instance
        """
        if idx in self.items:
            return self.items[idx]
        return None

    def get_all(self) -> Dict[int, NotepadItem]:
        """
        Get all notepads

        :return: notepads dict
        """
        return self.items

    def build(self) -> NotepadItem:
        """
        Build notepad

        :param idx: notepad idx
        :return: NotepadItem instance
        """
        item = NotepadItem()
        return item

    def add(self, notepad: NotepadItem) -> bool:
        """
        Add notepad

        :param notepad: NotepadItem instance
        :return: True if success
        """
        idx = self.provider.create(notepad)
        notepad.id = idx
        self.items[idx] = notepad
        self.save(idx)
        return True

    def update(self, notepad: NotepadItem) -> bool:
        """
        Update and save notepad

        :param notepad: NotepadItem instance
        :return: True if success
        """
        if notepad.idx not in self.items:
            return False

        notepad.updated_at = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
        self.items[notepad.idx] = notepad
        self.save(notepad.idx)
        return True

    def load(self, idx: int):
        """
        Load notepad by idx

        :param idx: notepad idx
        """
        self.items[idx] = self.provider.load(idx)

    def load_all(self):
        """Load all notepads"""
        self.items = self.provider.load_all()

    def save(self, idx: int) -> bool:
        """
        Save notepad by idx

        :param idx: notepad idx
        :return: True if saved, False if not
        :rtype: bool
        """
        if idx not in self.items:
            return False

        self.provider.save(self.items[idx])
        return False

    def save_all(self):
        """Save all notepads"""
        self.provider.save_all(self.items)

    def import_from_db(self) -> Optional[Dict[int, dict]]:
        """
        Import notepad tabs from database

        :return: dict with tabs
        """
        self.load_all()
        items = self.get_all()
        if len(items) == 0:
            return
        data = {}
        for idx in items:
            item = items[idx]
            tab = {
                "uuid": uuid.uuid4(),
                "pid": 0,
                "idx": item.idx,
                "type": Tab.TAB_NOTEPAD,
                "data_id": item.idx,
                "title": item.title,
            }
            if tab['title'] == "":
                tab['title'] = trans('output.tab.notepad')
            data[item.idx] = tab
        return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Optional, Dict, Any

from pygpt_net.core.types import (
    MODE_CHAT,
    MODE_COMPLETION,
)

from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.item.ctx import CtxItem

from .chat import Chat
from .completion import Completion


class Chain:
    def __init__(self, window=None):
        """
        Langchain wrapper core

        :param window: Window instance
        """
        self.window = window
        self.chat = Chat(window)
        self.completion = Completion(window)

    def call(
            self,
            context: BridgeContext,
            extra: Optional[Dict[str, Any]] = None
    ) -> bool:
        """
        Call LLM using Langchain

        :param context: Bridge context
        :param extra: Extra arguments
        """
        prompt = context.prompt
        system_prompt = context.system_prompt
        stream = context.stream
        model = context.model
        ctx = context.ctx

        if ctx is None:
            ctx = CtxItem()  # create empty context

        user_name = ctx.input_name  # from ctx
        ai_name = ctx.output_name  # from ctx
        response = None
        used_tokens = 0
        sub_mode = MODE_CHAT

        # get available sub-modes
        if 'mode' in model.langchain:
            if MODE_CHAT in model.langchain['mode']:
                sub_mode = MODE_CHAT
            elif MODE_COMPLETION in model.langchain['mode']:
                sub_mode = MODE_COMPLETION

        try:
            if sub_mode == MODE_CHAT:
                response = self.chat.send(
                    prompt=prompt,
                    system_prompt=system_prompt,
                    model=model,
                    history=context.history,
                    stream=stream,
                    ai_name=ai_name,
                    user_name=user_name,
                )
                used_tokens = self.chat.get_used_tokens()
            elif sub_mode == MODE_COMPLETION:
                response = self.completion.send(
                    prompt=prompt,
                    system_prompt=system_prompt,
                    model=model,
                    history=context.history,
                    stream=stream,
                    ai_name=ai_name,
                    user_name=user_name,
                )
                used_tokens = self.completion.get_used_tokens()

        except Exception as e:
            self.window.core.debug.log(e)
            raise e  # re-raise to window

        # if stream mode, store stream
        if stream:
            ctx.stream = response
            ctx.input_tokens = used_tokens  # get from input tokens calculation
            ctx.set_output("", ai_name)
            return True

        if response is None:
            return False

        # get output
        output = None
        if sub_mode == MODE_CHAT:
            output = response.content
        elif sub_mode == MODE_COMPLETION:
            output = response

        # store context
        ctx.set_output(output, ai_name)
        return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.18 19:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.config import BaseConfig, BasePlugin


class Config(BaseConfig):
    def __init__(self, plugin: BasePlugin = None, *args, **kwargs):
        super(Config, self).__init__(plugin)
        self.plugin = plugin

    def from_defaults(self, plugin: BasePlugin = None):
        """
        Set default options for plugin

        :param plugin: plugin instance
        """
        prompt = """
                # Controlling the Mouse and Keyboard:

                You have access to commands that allow you to control the mouse and keyboard on my computer. 
                Your objective is to perform tasks by precisely utilizing these commands. 

                ## Follow these instructions carefully to ensure accurate and effective actions:

                1. Task Analysis: When assigned a task, break it down into clear, sequential steps needed to accomplish the goal.        
                2. Screen Assessment: Use the provided screenshot to understand the current state of the screen. 
                Identify interactive elements such as buttons, menus, and text fields that are relevant to the task.        
                3. Command Utilization: Employ the available commands (e.g., mouse_move(x, y), mouse_click(button), 
                keyboard_type(text)) to interact with the identified elements. Ensure you use the correct syntax and parameters for each command.        
                4. Action Planning: Before executing commands, plan the exact coordinates and sequence of actions required. 
                This may involve calculating the position of an element based on the screenshot.        
                5. Step-by-Step Execution: Perform one command at a time. After each action, wait for the updated screenshot 
                to verify that the action had the intended effect before proceeding.        
                6. Verification and Adjustment: Use the feedback from each new screenshot to confirm successful execution. 
                If the outcome is not as expected, adjust your approach accordingly.        
                7. Focus on Relevant Areas: Begin by directing your actions to the area of the screen where the task is 
                to be performed. This may involve moving the cursor to a specific window or section.        
                8. Autonomous Completion: Complete the task independently using the commands provided. Avoid requesting 
                user intervention and ensure each step is necessary and contributes to the final goal.        
                9. Safety and Accuracy: Avoid any actions that may cause unintended side effects. Double-check coordinates 
                and command parameters to prevent errors.        
                10. Efficiency: Aim to accomplish the task in as few steps as possible while maintaining accuracy. 
                Consolidate actions when appropriate without skipping necessary verification.        
                11. Remember to focus on the correct window before typing on the keyboard.

                ## Example Workflow:

                Task: Find the window "App" and set the focus on it by clicking on it.        
                Step 1: Get current mouse position and current screenshot.        
                Step 2: Analyze the screenshot to locate the "App" window.        
                Step 3: Plan the mouse movement to the coordinates of the "App" window.        
                Step 4: Use mouse_move(x, y) to move the cursor to the window.        
                Step 5: Analyze the screenshot to confirm that the submission was successful.        
                Step 6: Execute mouse_click('left') to set the focus.        
                Step 7: Analyze the screenshot to confirm that the focus is set.        
                Step 8: If confirmation is visible, the task is complete. If not, try again, 
                without asking the user for permission.        

                ## Available Commands:

                mouse_move(x, y): Moves the cursor to the specified (x, y) screen coordinates.        
                mouse_click(button): Simulates a mouse click with the specified button ('left', 'right', or 'middle').        
                keyboard_type(text): Types the specified text string at the current cursor focus.        
                keyboard_key(text): Presses the keyboard key at the current cursor focus.        
                get_mouse_position(): Returns the current mouse cursor coordinates.        
                get_screenshot(): Provides an updated screenshot of the current screen.
                Note: Always ensure that your actions are context-aware and based on the most recent screenshot to 
                account for any changes on the screen.  

                ## Important: Use only one command at a time and wait for the updated screenshot before proceeding to the next step.

                Note: In case of failure, keep trying until the task is completed successfully.
                """
        plugin.add_option(
            "allow_mouse_move",
            type="bool",
            value=True,
            label="Allow mouse movement",
            description="Allow mouse movement",
        )
        plugin.add_option(
            "allow_mouse_click",
            type="bool",
            value=True,
            label="Allow mouse click",
            description="Allow mouse click",
        )
        plugin.add_option(
            "allow_mouse_scroll",
            type="bool",
            value=True,
            label="Allow mouse scroll",
            description="Allow mouse scroll",
        )
        plugin.add_option(
            "allow_keyboard",
            type="bool",
            value=True,
            label="Allow keyboard key press",
            description="Allow keyboard key press",
        )
        plugin.add_option(
            "allow_screenshot",
            type="bool",
            value=True,
            label="Allow making screenshot",
            description="Allow making screenshot",
        )
        plugin.add_option(
            "auto_focus",
            type="bool",
            value=False,
            label="Auto-focus on the window",
            description="Auto-focus on the window (mouse click) before keyboard typing",
        )
        plugin.add_option(
            "prompt",
            type="textarea",
            value=prompt,
            label="Prompt used to instruct how to control the mouse and keyboard",
            description="Prompt used to instruct how to control the mouse and keyboard",
            tooltip="Prompt used to instruct how to control the mouse and keyboard",
        )

        # commands
        plugin.add_cmd(
            "get_mouse_position",
            instruction="get current mouse coordinates and visual description of the screen provided by the vision model",
            params=[
                {
                    "name": "current_step",
                    "type": "text",
                    "description": "description of the current step in the task (for the vision model)",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: get current mouse position",
        )
        plugin.add_cmd(
            "get_screenshot",
            instruction="make screenshot",
            params=[],
            enabled=True,
            description="Enable: make screenshot",
        )
        plugin.add_cmd(
            "mouse_move",
            instruction="move mouse cursor to position (x, y)",
            params=[
                {
                    "name": "mouse_x",
                    "type": "int",
                    "description": "X coordinate",
                    "required": True,
                },
                {
                    "name": "mouse_y",
                    "type": "int",
                    "description": "Y coordinate",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: set mouse position to X,Y",
        )
        """
        plugin.add_cmd(
            "mouse_move",
            instruction="move mouse position by X offset, Y offset",
            params=[
                {
                    "name": "offset_x",
                    "type": "int",
                    "description": "offset X in pixels",
                    "required": True,
                },
                {
                    "name": "offset_y",
                    "type": "int",
                    "description": "offset Y in pixels",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: move mouse position by offset X,Y",
        )
        """
        plugin.add_cmd(
            "mouse_click",
            instruction="mouse button click",
            params=[
                {
                    "name": "button",
                    "type": "str",
                    "description": "button to click, enum: left|middle|right",
                    "required": True,
                },
                {
                    "name": "num_clicks",
                    "type": "int",
                    "description": "number of clicks, e.g. use 2 to double-click",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Mouse click",
        )
        plugin.add_cmd(
            "mouse_scroll",
            instruction="mouse scroll",
            params=[
                {
                    "name": "dx",
                    "type": "int",
                    "description": "steps to scroll horizontally, positive to the left, negative to the right",
                    "required": True,
                },
                {
                    "name": "dy",
                    "type": "int",
                    "description": "steps to scroll vertically, positive to scroll down, negative to scroll up",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Mouse scroll",
        )
        plugin.add_cmd(
            "keyboard_key",
            instruction="press keyboard key",
            params=[
                {
                    "name": "key",
                    "type": "str",
                    "description": "key to press",
                    "required": True,
                },
                {
                    "name": "modifier",
                    "type": "str",
                    "description": "modifier key, enum: shift|ctrl|alt|cmd",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Press keyboard keys",
        )
        plugin.add_cmd(
            "keyboard_type",
            instruction="type text on keyboard",
            params=[
                {
                    "name": "text",
                    "type": "str",
                    "description": "text to type on keyboard",
                    "required": True,
                },
                {
                    "name": "modifier",
                    "type": "str",
                    "description": "modifier key, enum: shift|ctrl|alt|cmd",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Type text on keyboard",
        )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.18 05:00:00                  #
# ================================================== #

from pygpt_net.ui.widget.dialog.snap import SnapDialogCamera, SnapDialogAudioInput, SnapDialogAudioOutput


class Snap:
    def __init__(self, window=None):
        """
        Snap dialogs

        :param window: Window instance
        """
        self.window = window

    def setup(self):
        """Setup snap dialog"""
        self.window.ui.dialog['snap_camera'] = SnapDialogCamera(self.window)
        self.window.ui.dialog['snap_audio_input'] = SnapDialogAudioInput(self.window)
        self.window.ui.dialog['snap_audio_output'] = SnapDialogAudioOutput(self.window)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Dict, Any

from pygpt_net.item.ctx import CtxItem
from pygpt_net.item.model import ModelItem

class BridgeContext:
    def __init__(self, **kwargs):
        """
        Bridge context

        :param kwargs: keyword arguments
        """
        self.ctx = kwargs.get("ctx", None)
        self.history = kwargs.get("history", [])
        self.mode = kwargs.get("mode", None)
        self.parent_mode = kwargs.get("parent_mode", None)  # real mode (global)
        self.model = kwargs.get("model", None)  # model instance, not model name
        self.temperature = kwargs.get("temperature", 1.0)
        self.prompt = kwargs.get("prompt", "")
        self.system_prompt = kwargs.get("system_prompt", "")
        self.system_prompt_raw = kwargs.get("system_prompt_raw", "")  # without plugins addons
        self.stream = kwargs.get("stream", False)
        self.assistant_id = kwargs.get("assistant_id", "")
        self.thread_id = kwargs.get("thread_id", "")
        self.external_functions = kwargs.get("external_functions", [])
        self.tools_outputs = kwargs.get("tools_outputs", [])
        self.max_tokens = kwargs.get("max_tokens", 150)
        self.idx = kwargs.get("idx", None)
        self.idx_mode = kwargs.get("idx_mode", "chat")
        self.attachments = kwargs.get("attachments", [])
        self.file_ids = kwargs.get("file_ids", [])
        self.reply_context = kwargs.get("reply_ctx", None)  # ReplyContext
        self.multimodal_ctx = kwargs.get("multimodal_ctx", MultimodalContext())  # AudioContext

        # check types
        if self.ctx is not None and not isinstance(self.ctx, CtxItem):
            raise ValueError("Invalid context instance")
        if self.model is not None and not isinstance(self.model, ModelItem):
            raise ValueError("Invalid model instance")

    def to_dict(self) -> Dict[str, Any]:
        """
        Return as dictionary

        :return: dictionary
        """
        data = {
            "ctx": self.ctx,
            "history": len(self.history),
            "mode": self.mode,
            "parent_mode": self.parent_mode,
            "model": self.model,
            "temperature": self.temperature,
            "prompt": self.prompt,
            "system_prompt": self.system_prompt,
            "system_prompt_raw": self.system_prompt_raw,
            "stream": self.stream,
            "assistant_id": self.assistant_id,
            "thread_id": self.thread_id,
            "external_functions": self.external_functions,
            "tools_outputs": self.tools_outputs,
            "max_tokens": self.max_tokens,
            "idx": self.idx,
            "idx_mode": self.idx_mode,
            "attachments": self.attachments,
            "file_ids": self.file_ids,
        }
        if self.ctx is not None:
            data["ctx"] = self.ctx.to_dict()
        if self.model is not None:
            data["model"] = self.model.to_dict()

        # sort by keys
        data = dict(sorted(data.items(), key=lambda item: item[0]))
        return data

class MultimodalContext:
    def __init__(self, **kwargs):
        """
        Multimodal context

        :param kwargs: keyword arguments
        """
        self.is_audio_input = kwargs.get("is_audio_input", False)
        self.is_audio_output = kwargs.get("is_audio_output", False)
        self.audio_data = kwargs.get("audio_data", None)
        self.audio_format = kwargs.get("audio_format", "wav")

    def to_dict(self) -> Dict[str, Any]:
        """
        Return as dictionary

        :return: dictionary
        """
        data = {
            "is_audio_input": self.is_audio_input,
            "is_audio_output": self.is_audio_output,
            "audio_format": self.audio_format,
        }
        # sort by keys
        data = dict(sorted(data.items(), key=lambda item: item[0]))
        return data


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.03.02 19:00:00                  #
# ================================================== #

from httpx_socks import SyncProxyTransport

from openai import OpenAI, DefaultHttpxClient

from pygpt_net.core.types import (
    MODE_ASSISTANT,
    MODE_AUDIO,
    MODE_CHAT,
    MODE_COMPLETION,
    MODE_IMAGE,
    MODE_VISION,
    MODE_RESEARCH,
)
from pygpt_net.core.bridge.context import BridgeContext

from .audio import Audio
from .assistants import Assistants
from .chat import Chat
from .completion import Completion
from .image import Image
from .store import Store
from .summarizer import Summarizer
from .vision import Vision


class Gpt:
    def __init__(self, window=None):
        """
        OpenAI API wrapper core

        :param window: Window instance
        """
        self.window = window
        self.assistants = Assistants(window)
        self.audio = Audio(window)
        self.chat = Chat(window)
        self.completion = Completion(window)
        self.image = Image(window)
        self.store = Store(window)
        self.summarizer = Summarizer(window)
        self.vision = Vision(window)

    def get_client(self, mode: str = MODE_CHAT) -> OpenAI:
        """
        Return OpenAI client

        :param mode: Mode
        :return: OpenAI client
        """
        args = {
            "api_key": self.window.core.config.get('api_key'),
            "organization": self.window.core.config.get('organization_key'),
        }
        # api endpoint
        if self.window.core.config.has('api_endpoint'):
            endpoint = self.window.core.config.get('api_endpoint')
            if endpoint:
                args["base_url"] = endpoint
        # proxy
        if self.window.core.config.has('api_proxy'):
            proxy = self.window.core.config.get('api_proxy')
            if proxy:
                transport = SyncProxyTransport.from_url(proxy)
                args["http_client"] = DefaultHttpxClient(
                    transport=transport,
                )

        # research mode endpoint - Perplexity
        if mode == MODE_RESEARCH:
            if self.window.core.config.has('api_key_perplexity'):
                args["api_key"] = self.window.core.config.get('api_key_perplexity')
            if self.window.core.config.has('api_endpoint_perplexity'):
                endpoint = self.window.core.config.get('api_endpoint_perplexity')
                if endpoint:
                    args["base_url"] = endpoint

        return OpenAI(**args)

    def call(self, context: BridgeContext, extra: dict = None) -> bool:
        """
        Call OpenAI API

        :param context: Bridge context
        :param extra: Extra arguments
        :return: result
        """
        mode = context.mode
        prompt = context.prompt
        stream = context.stream
        model = context.model  # model instance (item, not id)
        system_prompt = context.system_prompt
        assistant_id = context.assistant_id
        tools_outputs = context.tools_outputs
        max_tokens = context.max_tokens  # max output tokens

        ctx = context.ctx
        ai_name = ctx.output_name
        thread_id = ctx.thread  # from ctx

        # get model id
        model_id = None
        if model is not None:
            model_id = model.id
            if max_tokens > model.tokens:  # check max output tokens
                max_tokens = model.tokens

        response = None
        used_tokens = 0
        context.max_tokens = max_tokens  # update max output tokens
        file_ids = context.file_ids  # uploaded files IDs (assistant mode only)

        # completion
        if mode == MODE_COMPLETION:
            response = self.completion.send(
                context=context,
                extra=extra,
            )
            used_tokens = self.completion.get_used_tokens()

        # chat (OpenAI) | research (Perplexity)
        elif mode in [
            MODE_CHAT,
            MODE_AUDIO,
            MODE_RESEARCH
        ]:
            response = self.chat.send(
                context=context,
                extra=extra,
            )
            if hasattr(response, "citations"):
                if response.citations:
                    ctx.urls = response.citations
            used_tokens = self.chat.get_used_tokens()
            self.vision.append_images(ctx)  # append images to ctx if provided

        # image
        elif mode == MODE_IMAGE:
            return self.image.generate(
                context=context,
                extra=extra,
            )  # return here, async handled

        # vision
        elif mode == MODE_VISION:
            response = self.vision.send(
                context=context,
                extra=extra,
            )
            used_tokens = self.vision.get_used_tokens()
            self.vision.append_images(ctx)  # append images to ctx if provided

        # assistants
        elif mode == MODE_ASSISTANT:
            # check if assistant is already running and has tools outputs, then submit them, async handled
            if ctx.run_id is not None and len(tools_outputs) > 0:
                self.assistants.worker.tools_submit(
                    ctx,
                    model_id,
                    tools_outputs,  # list of tools outputs
                )
            else:
                # if not running, then send msg and create new assistant run, async handled
                self.assistants.worker.msg_send(
                    ctx,
                    thread_id,
                    assistant_id,
                    model_id,
                    file_ids,
                    prompt,
                    system_prompt,
                )
            return True  # if assistant mode then return here, will be handled async

        # if stream
        if stream:
            ctx.stream = response
            ctx.set_output("", ai_name)  # set empty output
            ctx.input_tokens = used_tokens  # get from input tokens calculation
            return True

        if response is None:
            return False

        # check for errors
        if "error" in response:
            print("Error in GPT response: " + str(response["error"]))
            return False

        # get output text from response (not-stream mode)
        output = ""
        if mode == MODE_COMPLETION:
            output = response.choices[0].text.strip()
        elif mode in [
            MODE_CHAT, 
            MODE_VISION, 
            MODE_RESEARCH
        ]:
            if response.choices[0]:
                if response.choices[0].message.content:
                    output = response.choices[0].message.content.strip()
                elif response.choices[0].message.tool_calls:
                    ctx.tool_calls = self.window.core.command.unpack_tool_calls(
                        response.choices[0].message.tool_calls,
                    )
        # audio
        elif mode in [MODE_AUDIO]:
            if response.choices[0]:
                if response.choices[0].message and response.choices[0].message.audio:
                    ctx.audio_output = response.choices[0].message.audio.data
                    ctx.audio_id = response.choices[0].message.audio.id
                    ctx.audio_expires_ts = response.choices[0].message.audio.expires_at
                    ctx.is_audio = True
                    output = response.choices[0].message.audio.transcript  # from transcript
                    self.chat.audio_prev_expires_ts = ctx.audio_expires_ts
                    self.chat.audio_prev_id = ctx.audio_id
                elif response.choices[0].message and not response.choices[0].message.audio:
                    output = response.choices[0].message.content
                    ctx.audio_id = self.chat.audio_prev_id
                    ctx.audio_expires_ts = self.chat.audio_prev_expires_ts
                if response.choices[0].message.tool_calls:
                    ctx.tool_calls = self.window.core.command.unpack_tool_calls(
                        response.choices[0].message.tool_calls,
                    )

        ctx.set_output(output, ai_name)
        ctx.set_tokens(
            response.usage.prompt_tokens,
            response.usage.completion_tokens,
        )
        return True

    def quick_call(self, context: BridgeContext, extra: dict = None) -> str:
        """
        Quick call OpenAI API with custom prompt

        :param context: Bridge context
        :param extra: Extra arguments
        :return: response content
        """
        mode = context.mode
        prompt = context.prompt
        system_prompt = context.system_prompt
        max_tokens = context.max_tokens
        temperature = context.temperature
        model = context.model
        if model is None:
            model = self.window.core.models.from_defaults()

        client = self.get_client(mode)
        messages = []
        messages.append({"role": "system", "content": system_prompt})
        messages.append({"role": "user", "content": prompt})
        additional_kwargs = {}
        if max_tokens > 0:
            additional_kwargs["max_tokens"] = max_tokens
        
        try:
            response = client.chat.completions.create(
                messages=messages,
                model=model.id,
                temperature=temperature,
                top_p=1.0,
                frequency_penalty=0.0,
                presence_penalty=0.0,
                **additional_kwargs,
            )
            return response.choices[0].message.content
        except Exception as e:
            self.window.core.debug.log(e)
            print("Error in GPT quick call: " + str(e))

    def stop(self):
        """Stop OpenAI API"""
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.08.25 04:00:00                  #
# ================================================== #

from .base_combo import BaseCombo

class LlamaModeCombo(BaseCombo):
    def __init__(self, window=None, parent_id: str = None, id: str = None, option: dict = None):
        """
        Llama index mode select combobox

        :param window: main window
        :param id: option id
        :param parent_id: parent option id
        :param option: option data
        """
        super(LlamaModeCombo, self).__init__(window, parent_id, id, option)

    def on_combo_change(self, index):
        """
        On combo change

        :param index: combo index
        """
        if not self.initialized:
            return
        self.current_id = self.combo.itemData(index)
        self.window.controller.idx.select_mode(self.current_id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.18 19:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.config import BaseConfig, BasePlugin


class Config(BaseConfig):
    def __init__(self, plugin: BasePlugin = None, *args, **kwargs):
        super(Config, self).__init__(plugin)
        self.plugin = plugin

    def from_defaults(self, plugin: BasePlugin = None):
        """
        Set default options for plugin

        :param plugin: plugin instance
        """
        plugin.add_option(
            "provider",
            type="combo",
            value="openai_whisper",
            label="Provider",
            description="Select audio transcribe provider, default: OpenAI Whisper",
            tooltip="Select audio transcribe provider",
            keys=plugin.get_provider_options(),
        )
        plugin.add_option(
            "auto_send",
            type="bool",
            value=True,
            label="Auto send",
            description="Automatically send input when voice is detected. Default: True",
        )
        plugin.add_option(
            "advanced",
            type="bool",
            value=False,
            label="Enable advanced mode",
            description="Enable only if you want to use advanced mode and settings below, "
                        "do not enable this option if you just want to use simple mode (default)",
        )
        plugin.add_option(
            "timeout",
            type="int",
            value=5,
            label="Timeout",
            description="Speech recognition timeout. Default: 5",
            min=0,
            max=30,
            slider=True,
            tooltip="Timeout, default: 5",
            advanced=True,
        )
        plugin.add_option(
            "phrase_length",
            type="int",
            value=10,
            label="Phrase max length",
            description="Speech recognition phrase length. Default: 10",
            min=0,
            max=30,
            slider=True,
            tooltip="Phrase max length, default: 10",
            advanced=True,
        )
        plugin.add_option(
            "min_energy",
            type="float",
            value=1.3,
            label="Min. energy",
            description="Minimum threshold multiplier above the noise level to begin recording;"
                        " 1 = disabled. Default: 1.3",
            min=1,
            max=50,
            slider=True,
            tooltip="Min. energy, default: 1.3, 1 = disabled, adjust for your microphone",
            multiplier=10,
            advanced=True,
        )
        plugin.add_option(
            "adjust_noise",
            type="bool",
            value=True,
            label="Adjust ambient noise",
            description="Adjust for ambient noise. Default: True",
            advanced=True,
        )
        plugin.add_option(
            "continuous_listen",
            type="bool",
            value=False,
            label="Continuous listening",
            description="EXPERIMENTAL: continuous listening - do not stop listening after a single input.\n"
                        "Warning: This feature may lead to unexpected results and requires fine-tuning "
                        "with the rest of the options!",
            advanced=True,
        )
        plugin.add_option(
            "wait_response",
            type="bool",
            value=True,
            label="Wait for response",
            description="Wait for a response before listening for the next input. Default: True",
            advanced=True,
        )
        plugin.add_option(
            "magic_word",
            type="bool",
            value=False,
            label="Magic word",
            description="Activate listening only after the magic word is provided, "
                        "like 'Hey GPT' or 'OK GPT'. Default: False",
            advanced=True,
        )
        plugin.add_option(
            "magic_word_reset",
            type="bool",
            value=True,
            label="Reset Magic word",
            description="Reset the magic word status after it is received "
                        "(the magic word will need to be provided again). Default: True",
            advanced=True,
        )
        plugin.add_option(
            "magic_words",
            type="text",
            value="OK, Okay, Hey GPT, OK GPT",
            label="Magic words",
            description="Specify magic words for 'Magic word' option: if received this word then "
                        "start listening, put words separated by comma. Magic word option must be enabled, "
                        "examples: \"Hey GPT, OK GPT\"",
            advanced=True,
        )
        plugin.add_option(
            "magic_word_timeout",
            type="int",
            value=1,
            label="Magic word timeout",
            description="Magic word recognition timeout. Default: 1",
            min=0,
            max=30,
            slider=True,
            tooltip="Timeout, default: 1",
            advanced=True,
        )
        plugin.add_option(
            "magic_word_phrase_length",
            type="int",
            value=2,
            label="Magic word phrase max length",
            description="Magic word phrase length. Default: 2",
            min=0,
            max=30,
            slider=True,
            tooltip="Phrase length, default: 2",
            advanced=True,
        )
        plugin.add_option(
            "prefix_words",
            type="text",
            value="",
            label="Prefix words",
            description="Specify prefix words: if defined, only phrases starting with these words "
                        "will be transmitted, and the remainder will be ignored. Separate the words with "
                        "a comma., eg. 'OK, Okay, GPT'. Leave empty to disable",
            advanced=True,
        )
        plugin.add_option(
            "stop_words",
            type="text",
            value="stop, exit, quit, end, finish, close, terminate, kill, halt, abort",
            label="Stop words",
            description="Specify stop words: if any of these words are received, then stop listening. "
                        "Separate the words with a comma, or leave it empty to disable the feature, "
                        "default: stop, exit, quit, end, finish, close, terminate, kill, halt, abort",
            advanced=True,
        )

        # advanced options
        plugin.add_option(
            "recognition_energy_threshold",
            type="int",
            value=300,
            label="energy_threshold",
            description="Represents the energy level threshold for sounds. Default: 300",
            min=0,
            max=10000,
            slider=True,
            advanced=True,
        )
        plugin.add_option(
            "recognition_dynamic_energy_threshold",
            type="bool",
            value=True,
            label="dynamic_energy_threshold",
            description="Represents whether the energy level threshold for sounds should be automatically "
                        "adjusted based on the currently ambient noise level while listening. "
                        "Default: True",
            advanced=True,
        )
        plugin.add_option(
            "recognition_dynamic_energy_adjustment_damping",
            type="float",
            value=0.15,
            label="dynamic_energy_adjustment_damping",
            description="Represents approximately the fraction of the current energy threshold that is "
                        "retained after one second of dynamic threshold adjustment. Default: 0.15",
            min=0,
            max=100,
            slider=True,
            multiplier=100,
            advanced=True,
        )
        plugin.add_option(
            "recognition_pause_threshold",
            type="float",
            value=0.8,
            label="pause_threshold",
            description="Represents the minimum length of silence (in seconds) that will register as the end "
                        "of a phrase.\nDefault: 0.8",
            min=0,
            max=100,
            slider=True,
            multiplier=10,
            advanced=True,
        )
        plugin.add_option(
            "recognition_adjust_for_ambient_noise_duration",
            type="float",
            value=1,
            label="adjust_for_ambient_noise: duration",
            description="The duration parameter is the maximum number of seconds that it will dynamically "
                        "adjust the threshold for before returning. Default: 1",
            min=0,
            max=100,
            slider=True,
            multiplier=10,
            advanced=True,
        )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.08 17:00:00                  #
# ================================================== #

from datetime import datetime, timedelta

from pygpt_net.utils import trans


class Helpers:
    def __init__(self, window=None):
        """
        Helpers - voice get data

        :param window: Window instance
        """
        self.window = window

    def get_status(self) -> str:
        """
        Get the current application status

        :return: status, text to read
        """
        ctx = ""
        total = self.window.core.ctx.count_meta()
        meta = self.window.core.ctx.get_current_meta()
        mode = trans("mode." + self.window.core.config.get("mode"))
        mode += " " + self.get_selected_model()
        mode += " " + self.get_selected_preset()
        tab = self.window.controller.ui.tabs.get_current_tab_name()
        last = ""
        msg = ""
        if meta is not None:
            ctx = meta.name
            last = self.convert_date(meta.updated)
        try:
            msg = trans("event.audio.app.status").format(
                mode=mode,
                tab=tab,
                ctx=ctx,
                last=last,
                total=total,
            )
        except Exception as e:
            pass
        return msg

    def get_current_ctx(self) -> str:
        """
        Get current context

        :return: text to read
        """
        meta = self.window.core.ctx.get_current_meta()
        if meta is not None:
            name = meta.name
            updated = self.convert_date(meta.updated)
            try:
                msg = trans("event.audio.ctx.current").format(
                    ctx=name,
                    last=updated,
                )
            except Exception as e:
                msg = ""
            return msg
        return ""

    def get_selected_ctx(self) -> str:
        """
        Get current context

        :return: text to read
        """
        meta = self.window.core.ctx.get_current_meta()
        if meta is not None:
            name = meta.name
            updated = self.convert_date(meta.updated)
            try:
                msg = trans("event.audio.ctx.selected").format(
                    ctx=name,
                    last=updated,
                )
            except Exception as e:
                msg = ""
            return msg
        return ""

    def get_selected_tab(self) -> str:
        """
        Get current tab

        :return: text to read
        """
        msg = ""
        tab = self.window.controller.ui.tabs.get_current_tab_name_for_audio()
        try:
            msg = trans("event.audio.tab.switch").format(
                tab=tab,
            )
        except Exception as e:
            pass
        return msg

    def get_selected_mode(self) -> str:
        """
        Get current mode

        :return: text to read
        """
        mode = self.window.core.config.get("mode")
        if mode is None:
            return ""
        try:
            msg = trans("event.audio.mode.selected").format(
                mode=trans("mode." + mode),
            )
        except Exception as e:
            msg = ""
        return msg

    def get_selected_model(self) -> str:
        """
        Get current model

        :return: text to read
        """
        model = self.window.core.config.get("model")
        if model is None:
            return ""
        try:
            msg = trans("event.audio.model.selected").format(
                model=model,
            )
        except Exception as e:
            msg = ""
        return msg

    def get_selected_preset(self) -> str:
        """
        Get current preset

        :return: text to read
        """
        preset_id = self.window.core.config.get("preset")
        if preset_id is None:
            return ""
        mode = self.window.core.config.get("mode")
        preset = self.window.core.presets.get_by_id(mode, preset_id)
        if preset is None:
            return ""
        try:
            msg = trans("event.audio.preset.selected").format(
                preset=preset.name,
            )
        except Exception as e:
            msg = ""
        return msg

    def get_last_ctx_item(self) -> str:
        """
        Get last context item

        :return: text to read
        """
        text = ""
        item = self.window.core.ctx.get_last()
        if item is not None:
            try:
                text = trans("event.audio.ctx.last").format(
                    input=item.input,
                    output=item.output,
                )
            except Exception as e:
                pass
        return text

    def get_all_ctx_items(self) -> str:
        """
        Get all context items

        :return: text to read
        """
        text = ""
        entries = []
        items = self.window.core.ctx.get_all_items(ignore_first=False)
        for item in items:
            try:
                entries.append(trans("event.audio.ctx.last").format(
                    input=item.input,
                    output=item.output,
                ))
            except Exception as e:
                pass
        if len(entries) > 0:
            text = "\n".join(entries)
        return text

    def convert_date(self, timestamp: int) -> str:
        """
        Convert timestamp to human readable format

        :param timestamp: timestamp
        :return: string
        """
        today = datetime.today().date()
        yesterday = today - timedelta(days=1)
        date = datetime.fromtimestamp(timestamp).date()
        hour_min = datetime.fromtimestamp(timestamp).strftime("%H:%M")

        days_ago = (today - date).days
        weeks_ago = days_ago // 7

        if date == today:
            return trans('dt.today') + " " + hour_min
        elif date == yesterday:
            return trans('dt.yesterday') + " " + hour_min
        elif weeks_ago == 1:
            return trans('dt.week')
        elif 1 < weeks_ago < 4:
            return f"{weeks_ago} " + trans('dt.weeks')
        elif days_ago < 30:
            return f"{days_ago} " + trans('dt.days_ago')
        elif 30 <= days_ago < 32:
            return trans('dt.month')
        else:
            return date.strftime("%Y-%m-%d")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

import os
from typing import Optional, List

from pygpt_net.item.assistant import AssistantStoreItem


class Store:
    def __init__(self, window=None):
        """
        GPT vector stores API Wrapper

        :param window: Window instance
        """
        self.window = window

    def get_client(self):
        """
        Get OpenAI client

        :return: OpenAI client
        """
        return self.window.core.gpt.get_client()

    def log(
            self,
            msg: str,
            callback: Optional[callable] = None
    ):
        """
        Log message

        :param msg: message to log
        :param callback: callback log function
        """
        if callback is not None:
            callback(msg)
        else:
            print(msg)

    def get_file(self, file_id: str):
        """
        Get file info

        :param file_id: file ID
        :return: file info
        """
        client = self.get_client()
        return client.files.retrieve(file_id)

    def upload(
            self,
            path: str,
            purpose: str = "assistants"
    ) -> Optional[str]:
        """
        Upload file to assistant

        :param path: file path
        :param purpose: file purpose
        :return: file ID or None
        """
        client = self.get_client()
        if not os.path.exists(path):
            return None

        result = client.files.create(
            file=open(path, "rb"),
            purpose=purpose,
        )
        if result is not None:
            return result.id

    def download(
            self,
            file_id: str,
            path: str
    ):
        """
        Download file

        :param file_id: file ID
        :param path: path to save file
        """
        client = self.window.core.gpt.get_client()
        content = client.files.content(file_id)
        data = content.read()
        with open(path, 'wb', ) as f:
            f.write(data)
            
    def get_files_ids_all(
            self,
            items: list,
            order: str = "asc",
            limit: int = 100,
            after: Optional[str] = None,
    ) -> list:
        """
        Get all vector store IDs

        :param items: items
        :param order: order
        :param limit: limit
        :param after: next page after ID
        :return: files IDs (all, paginated)
        """
        client = self.get_client()
        args = {
            "order": order,
            "limit": limit,
        }
        if after is not None:
            args['after'] = after
        stores = client.files.list(**args)
        if stores is not None:
            for remote in stores.data:
                id = remote.id
                if id not in items:
                    items.append(id)
            # next page
            if stores.has_more:
                return self.get_files_ids_all(
                    items,
                    order,
                    limit,
                    stores.last_id,
                )
        return items

    def get_files_ids(self) -> List[str]:
        """
        Get all files IDs

        :return: files IDs
        """
        client = self.get_client()
        items = []
        stores = client.files.list()
        if stores is not None:
            for remote in stores.data:
                id = remote.id
                if id not in items:
                    items.append(id)
        return items

    def remove_files(
            self,
            callback: Optional[callable] = None
    ) -> int:
        """
        Remove all files

        :param callback: callback function
        :return: number of deleted files
        """
        num = 0
        files = self.get_files_ids()
        for file_id in files:
            self.log("Removing file: " + file_id, callback)
            try:
                self.delete_file(file_id)
                num += 1
            except Exception as e:
                msg = "Error removing file {}: {}".format(file_id, str(e))
                self.log(msg, callback)
        return num

    def remove_store_files(
            self,
            store_id: str,
            callback: Optional[callable] = None
    ) -> int:
        """
        Remove all files from store

        :param store_id: store ID
        :param callback: callback function
        :return: number of deleted files
        """
        num = 0
        files = self.get_store_files_ids(store_id, [])
        for file_id in files:
            self.log("Removing file: " + file_id, callback)
            try:
                self.delete_store_file(store_id, file_id)
                num += 1
            except Exception as e:
                msg = "Error removing file {}: {}".format(file_id, str(e))
                self.log(msg, callback)
        return num

    def import_stores(
            self,
            items: dict,
            order: str = "asc",
            limit: int = 100,
            after: Optional[str] = None,
            callback: Optional[callable] = None
    ) -> dict:
        """
        Import vector stores from API

        :param items: items
        :param order: order
        :param limit: limit
        :param after: next page after ID
        :param callback: callback function
        :return: vector stores objects
        """
        client = self.get_client()
        args = {
            "order": order,
            "limit": limit,
        }
        if after is not None:
            args['after'] = after
        stores = client.beta.vector_stores.list(**args)
        if stores is not None:
            for remote in stores.data:
                id = remote.id
                if id not in items:
                    items[id] = AssistantStoreItem()
                tmp_name = remote.name
                if tmp_name is None:
                    items[id].is_thread = True  # tmp store for thread
                    tmp_name = ""
                items[id].id = id
                items[id].name = tmp_name
                items[id].file_ids = []
                items[id].status = self.window.core.assistants.store.parse_status(remote)
                self.window.core.assistants.store.append_status(items[id], items[id].status)
                self.log("Imported vector store: " + id, callback)
            # next page
            if stores.has_more:
                return self.import_stores(
                    items,
                    order,
                    limit,
                    stores.last_id,
                    callback,
                )
        return items

    def create_store(
            self,
            name: str,
            expire_days: int = 0
    ):
        """
        Create vector store

        :param name: store name
        :param expire_days: expire days
        :return: vector store object
        """
        client = self.get_client()
        expires_after = {
            "anchor": "last_active_at",
            "days": expire_days,
        }
        if expire_days <= 0:
            expires_after = None
        vector_store = client.beta.vector_stores.create(
            name=name,
            expires_after=expires_after,
        )
        if vector_store is not None:
            return vector_store

    def update_store(
            self,
            id: str,
            name: str,
            expire_days: int = 0
    ):
        """
        Update vector store

        :param id: store id
        :param name: store name
        :param expire_days: expire days
        :return: vector store object
        """
        client = self.get_client()
        expires_after = {
            "anchor": "last_active_at",
            "days": expire_days,
        }
        if expire_days <= 0:
            expires_after = None
        vector_store = client.beta.vector_stores.update(
            vector_store_id=id,
            name=name,
            expires_after=expires_after,
        )
        if vector_store is not None:
            return vector_store

    def get_store(self, id: str):
        """
        Get vector store by ID

        :param id: store id
        :return: vector store
        """
        client = self.get_client()
        vector_store = client.beta.vector_stores.retrieve(
            vector_store_id=id,
        )
        if vector_store is not None:
            return vector_store

    def remove_store(self, id: str):
        """
        Delete vector store by ID

        :param id: store id
        :return: vector store object
        """
        client = self.get_client()
        vector_store = client.beta.vector_stores.delete(
            vector_store_id=id,
        )
        if vector_store is not None:
            return vector_store

    def get_stores_ids(
            self,
            items: list,
            order: str = "asc",
            limit: int = 100,
            after: Optional[str] = None,
    ) -> list:
        """
        Get all vector stores IDs

        :param items: items
        :param order: order
        :param limit: limit
        :param after: next page after ID
        :return: IDs dict (all, paginated)
        """
        client = self.get_client()
        args = {
            "order": order,
            "limit": limit,
        }
        if after is not None:
            args['after'] = after
        stores = client.beta.vector_stores.list(**args)
        if stores is not None:
            for remote in stores.data:
                id = remote.id
                if id not in items:
                    items.append(id)
            # next page
            if stores.has_more:
                return self.get_stores_ids(
                    items,
                    order,
                    limit,
                    stores.last_id,
                )
        return items

    def get_store_files_ids(
            self,
            store_id: str,
            items: list,
            order: str = "asc",
            limit: int = 100,
            after: Optional[str] = None,
    ) -> list:
        """
        Get all vector store files IDs

        :param store_id: store ID
        :param items: items
        :param order: order
        :param limit: limit
        :param after: next page after ID
        :return: IDs dict  (all, paginated)
        """
        client = self.get_client()
        args = {
            "vector_store_id": store_id,
            "order": order,
            "limit": limit,
        }
        if after is not None:
            args['after'] = after
        files = client.beta.vector_stores.files.list(**args)
        if files is not None:
            for remote in files.data:
                id = remote.id
                if id not in items:
                    items.append(id)
            # next page
            if files.has_more:
                return self.get_store_files_ids(
                    store_id,
                    items,
                    order,
                    limit,
                    files.last_id,
                )
        return items

    def remove_from_stores(self) -> int:
        """
        Remove all files from vector stores

        :return: number of deleted files
        """
        stores = self.get_stores_ids([])
        num = 0
        for store_id in stores:
            files = self.get_store_files_ids(store_id, [])
            for file_id in files:
                self.log("Removing file from vector store [{}]:{} ".format(store_id, file_id))
                self.delete_store_file(store_id, file_id)
                num += 1
        return num

    def remove_from_store(self, store_id: str) -> int:
        """
        Remove all files from vector store

        :param store_id: store ID
        :return: number of deleted files
        """
        files = self.get_store_files_ids(store_id, [])
        num = 0
        for file_id in files:
            self.log("Removing file from vector store [{}]:{} ".format(store_id, file_id))
            self.delete_store_file(store_id, file_id)
            num += 1
        return num

    def remove_all(
            self,
            callback: Optional[callable] = None
    ) -> int:
        """
        Remove all vector stores

        :param callback: callback function
        :return: number of deleted stores
        """
        num = 0
        stores = self.get_stores_ids([])
        for store_id in stores:
            self.log("Removing vector store: " + store_id, callback)
            try:
                self.remove_store(store_id)
                num += 1
            except Exception as e:
                msg = "Error removing vector store {}: {}".format(store_id, str(e))
                self.log(msg, callback)
        return num

    def add_file(
            self,
            store_id: str,
            file_id: str
    ):
        """
        Add file to vector store

        :param store_id: store id
        :param file_id: file id
        :return: vector store file
        """
        client = self.get_client()
        vector_store_file = client.beta.vector_stores.files.create(
            vector_store_id=store_id,
            file_id=file_id,
        )
        if vector_store_file is not None:
            return vector_store_file

    def delete_file(
            self,
            file_id: str
    ) -> str:
        """
        Delete file from assistant

        :param file_id: file ID
        :return: file ID
        """
        client = self.get_client()
        deleted_file = client.files.delete(
            file_id=file_id
        )
        if deleted_file is not None:
            if deleted_file is not None:
                return deleted_file.id

    def delete_store_file(
            self,
            store_id: str,
            file_id: str
    ):
        """
        Delete file from vector store

        :param store_id: store id
        :param file_id: file id
        :return: vector store file
        """
        client = self.get_client()
        vector_store_file = client.beta.vector_stores.files.delete(
            vector_store_id=store_id,
            file_id=file_id,
        )
        if vector_store_file is not None:
            return vector_store_file

    def import_stores_files(
            self,
            callback: Optional[callable] = None
    ) -> int:
        """
        Import all vector stores files

        :param callback: callback function
        :return: num of imported files
        """
        store_ids = self.get_stores_ids([])
        num = 0
        for store_id in store_ids:
            items = []
            try:
                items = self.import_store_files(
                    store_id,
                    items,
                    callback=callback,
                )
            except Exception as e:
                msg = "Error importing store {} files list: {}".format(store_id, str(e))
                self.log(msg, callback)
            num += len(items)
        return num

    def import_store_files(
            self,
            store_id: str,
            items: list,
            order: str = "asc",
            limit: int = 100,
            after: str = None,
            callback: Optional[callable] = None
    ) -> list:
        """
        Import and get all vector store files IDs

        :param store_id: store ID
        :param items: items
        :param order: order
        :param limit: limit
        :param after: next page after ID
        :param callback: callback function
        :return: imported IDs dict
        """
        client = self.get_client()
        args = {
            "vector_store_id": store_id,
            "order": order,
            "limit": limit,
        }
        if after is not None:
            args['after'] = after
        files = client.beta.vector_stores.files.list(**args)
        if files is not None:
            for remote in files.data:
                try:
                    id = remote.id
                    if id not in items:
                        items.append(id)
                        data = self.get_file(remote.id)
                        self.window.core.assistants.files.insert(store_id, data)  # add remote file to DB
                        msg = "Imported file ID {} to store {}".format(remote.id, store_id)
                        self.log(msg, callback)
                except Exception as e:
                    msg = "Error importing file {} to store {}: {}".format(remote.id, store_id, str(e))
                    self.log(msg, callback)
            # next page
            if files.has_more:
                return self.import_store_files(
                    store_id,
                    items,
                    order,
                    limit,
                    files.last_id,
                    callback,
                )
        return items



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QLabel, QWidget, QVBoxLayout

from pygpt_net.utils import trans


class VideoContainer(QWidget):
    def __init__(self, window=None):
        """
        Video container with scroll

        :param window: Window instance
        """
        super(VideoContainer, self).__init__()
        self.window = window
        self.setStyleSheet("background-color: #000000;")

        self.label = QLabel(trans("vision.capture.label"))
        self.label.setMaximumHeight(15)
        self.label.setAlignment(Qt.AlignCenter)

        self.layout = QVBoxLayout()
        self.video = VideoLabel(window=self.window)
        self.video.setStyleSheet("background-color: #000000;")
        self.layout.addWidget(self.video)
        self.layout.addWidget(self.label)
        self.setLayout(self.layout)


class VideoLabel(QLabel):
    def __init__(self, text=None, window=None):
        """
        Video output label

        :param text: text
        :param window: main window
        """
        super(VideoLabel, self).__init__(text)
        self.window = window

    def mousePressEvent(self, event):
        """
        Mouse click

        :param event: mouse event
        """
        if event.button() == Qt.LeftButton:
            self.window.controller.camera.manual_capture()
        elif event.button() == Qt.RightButton:
            pass
        elif event.button() == Qt.MiddleButton:
            pass

    def mouseDoubleClickEvent(self, event):
        """
        Mouse double click

        :param event: mouse event
        """
        if event.button() == Qt.LeftButton:
            pass
        elif event.button() == Qt.RightButton:
            pass
        elif event.button() == Qt.MiddleButton:
            pass

    def mouseReleaseEvent(self, event):
        """
        Mouse release

        :param event: mouse event
        """
        if event.button() == Qt.LeftButton:
            pass
        elif event.button() == Qt.RightButton:
            pass
        elif event.button() == Qt.MiddleButton:
            pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

import json
import os
import re
from datetime import datetime
from typing import Optional, List

from pygpt_net.core.render.base import BaseRenderer
from pygpt_net.core.text.utils import has_unclosed_code_tag
from pygpt_net.item.ctx import CtxItem, CtxMeta
from pygpt_net.ui.widget.textarea.input import ChatInput
from pygpt_net.utils import trans
from pygpt_net.core.tabs.tab import Tab

from .body import Body
from .helpers import Helpers
from .parser import Parser
from .pid import PidData


class Renderer(BaseRenderer):
    NODE_INPUT = 0
    NODE_OUTPUT = 1

    def __init__(self, window=None):
        super(Renderer, self).__init__(window)
        """
        WebEngine renderer

        :param window: Window instance
        """
        self.window = window
        self.body = Body(window)
        self.helpers = Helpers(window)
        self.parser = Parser(window)
        self.pids = {}  # per node data

    def prepare(self):
        """
        Prepare renderer
        """
        self.pids = {}

    def on_load(self, meta: CtxMeta = None):
        """
        On load (meta) event

        :param meta: context meta
        """
        node = self.get_output_node(meta)
        node.set_meta(meta)
        self.reset(meta)

    def on_page_loaded(
            self,
            meta: Optional[CtxMeta] = None,
            tab: Optional[Tab] = None
    ):
        """
        On page loaded callback from WebEngine widget

        :param meta: context meta
        :param tab: Tab
        """
        if meta is None:
            return
        pid = tab.pid
        if pid is None or pid not in self.pids:
            return
        self.pids[pid].loaded = True
        if self.pids[pid].html != "" and not self.pids[pid].use_buffer:
            self.clear_chunks_input(pid)
            self.clear_chunks_output(pid)
            self.clear_nodes(pid)
            self.append(pid, self.pids[pid].html, flush=True)
            self.pids[pid].html = ""

    def get_pid(self, meta: CtxMeta):
        """
        Get PID for context meta
        
        :param meta: context PID
        """
        return self.window.core.ctx.output.get_pid(meta)

    def get_or_create_pid(self, meta: CtxMeta):
        """
        Get PID for context meta and create PID data (if not exists)
        
        :param meta: context PID
        """
        if meta is not None:
            pid = self.get_pid(meta)
            if pid not in self.pids:
                self.pid_create(pid, meta)
            return pid

    def pid_create(
            self,
            pid: Optional[int],
            meta: CtxMeta
    ):
        """
        Create PID data
        
        :param pid: PID
        :param meta: context meta
        """
        if pid is not None:
            self.pids[pid] = PidData(pid, meta)

    def init(self, pid: Optional[int]):
        """
        Initialize renderer

        :param pid: context PID
        """
        if pid in self.pids and not self.pids[pid].initialized:
            self.flush(pid)
            self.pids[pid].initialized = True
        else:
            self.clear_chunks(pid)

    def begin(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            stream: bool = False
    ):
        """
        Render begin

        :param meta: context meta
        :param ctx: context item
        :param stream: True if it is a stream
        """
        pid = self.get_or_create_pid(meta)
        self.init(pid)
        self.reset_names(meta)
        self.tool_output_end()  # reset tools

    def end(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            stream: bool = False
    ):
        """
        Render end

        :param meta: context meta
        :param ctx: context item
        :param stream: True if it is a stream
        """
        pid = self.get_or_create_pid(meta)
        if pid is None:
            return
        if self.pids[pid].item is not None and stream:
            self.append_context_item(meta, self.pids[pid].item)
            self.pids[pid].item = None

    def end_extra(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            stream: bool = False
    ):
        """
        Render end extra
        
        :param meta: context meta
        :param ctx: context item
        :param stream: True if it is a stream
        """
        self.to_end(ctx)

    def stream_begin(
            self,
            meta: CtxMeta,
            ctx: CtxItem
    ):
        """
        Render stream begin
        
        :param meta: context meta
        :param ctx: context item
        """
        pass

    def stream_end(
            self,
            meta: CtxMeta,
            ctx: CtxItem
    ):
        """
        Render stream end
        
        :param meta: context meta
        :param ctx: context item
        """
        pid = self.get_or_create_pid(meta)
        if self.window.controller.agent.legacy.enabled():
            if self.pids[pid].item is not None:
                self.append_context_item(meta, self.pids[pid].item)
                self.pids[pid].item = None

    def append_context(
            self,
            meta: CtxMeta,
            items: List[CtxItem],
            clear: bool = True
    ):
        """
        Append all context to output
        
        :param meta: Context meta
        :param items: context items
        :param clear: True if clear all output before append
        """
        if len(items) == 0:
            if meta is None:
                return

        pid = self.get_or_create_pid(meta)
        self.init(pid)

        if clear:
            self.reset(meta)
        i = 0

        self.pids[pid].use_buffer = True
        self.pids[pid].html = ""
        prev_ctx = None
        for item in items:
            self.update_names(meta, item)
            item.idx = i
            if i == 0:
                item.first = True
            next_item = items[i + 1] if i + 1 < len(items) else None   # append next item if exists
            self.append_context_item(
                meta,
                item,
                prev_ctx=prev_ctx,
                next_ctx=next_item
            )  # to html buffer
            prev_ctx = item
            i += 1
        self.pids[pid].use_buffer = False

        # flush
        if self.pids[pid].html != "":
            self.append(
                pid,
                self.pids[pid].html,
                flush=True
            )  # flush buffer if page loaded, otherwise it will be flushed on page load

    def append_input(
            self, meta: CtxMeta,
            ctx: CtxItem,
            flush: bool = True,
            append: bool = False
    ):
        """
        Append text input to output

        :param meta: context meta
        :param ctx: context item
        :param flush: flush HTML
        :param append: True if force append node
        """
        self.tool_output_end()  # reset tools
        pid = self.get_or_create_pid(meta)
        if not flush:
            self.clear_chunks_input(pid)

        self.update_names(meta, ctx)
        if ctx.input is None or ctx.input == "":
            return

        text = ctx.input

        # hidden internal call
        if ctx.internal \
                and not ctx.first \
                and not ctx.input.strip().startswith("user: ") \
                and not ctx.input.strip().startswith("@"):  # expert says:
            return
        else:
            # don't show user prefix if provided in internal call goal update
            if ctx.internal and ctx.input.startswith("user: "):
                text = re.sub(r'^user: ', '> ', ctx.input)

        if flush:  # to chunk buffer
            if self.is_stream() and not append:
                content = self.prepare_node(meta, ctx, text.strip(), self.NODE_INPUT)
                self.append_chunk_input(meta, ctx, content, False)
                return

        self.append_node(meta, ctx, text.strip(), self.NODE_INPUT)

    def append_output(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            flush: bool = True,
            prev_ctx: Optional[CtxItem] = None,
            next_ctx: Optional[CtxItem] = None
    ):
        """
        Append text output to output
        
        :param meta: context meta
        :param ctx: context item
        :param flush: flush HTML
        :param prev_ctx: previous context
        :param next_ctx: next context
        """
        self.tool_output_end()  # reset tools
        if ctx.output is None or ctx.output == "":
            return
        self.append_node(
            meta=meta,
            ctx=ctx,
            html=ctx.output.strip(),
            type=self.NODE_OUTPUT,
            prev_ctx=prev_ctx,
            next_ctx=next_ctx
        )

    def append_chunk(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            text_chunk: str,
            begin: bool = False
    ):
        """
        Append output chunk to output
        
        :param meta: context meta
        :param ctx: context item
        :param text_chunk: text chunk
        :param begin: if it is the beginning of the text
        """
        pid = self.get_or_create_pid(meta)
        self.pids[pid].item = ctx
        if text_chunk is None or text_chunk == "":
            if begin:
                self.pids[pid].buffer = ""  # always reset buffer
            return

        self.update_names(meta, ctx)
        raw_chunk = str(text_chunk)
        if begin:
            # debug
            debug = ""
            if self.is_debug():
                debug = self.append_debug(ctx, pid, "stream")
            if debug:
                raw_chunk = debug + raw_chunk
            self.pids[pid].buffer = ""  # reset buffer
            self.pids[pid].is_cmd = False  # reset command flag
            self.clear_chunks_output(pid)
        self.pids[pid].buffer += raw_chunk

        """
        # cooldown (throttling) to prevent high CPU usage on huge text chunks
        if len(self.buffer) > self.throttling_min_chars:
            current_time = time.time()
            if current_time - self.last_time_called <= self.cooldown:
                return  # wait a moment
            else:
                self.last_time_called = current_time
        """

        # parse chunks
        to_append = self.pids[pid].buffer
        if has_unclosed_code_tag(self.pids[pid].buffer):
            to_append += "\n```"  # fix for code block without closing ```
        html = self.parser.parse(to_append)
        escaped_chunk = json.dumps(html)
        try:
            self.get_output_node(meta).page().runJavaScript(
                f"replaceOutput('{self.pids[pid].name_bot}', {escaped_chunk});")
        except Exception as e:
            pass

    def append_chunk_input(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            text_chunk: str,
            begin: bool = False
    ):
        """
        Append output chunk to output

        :param meta: context meta
        :param ctx: context item
        :param text_chunk: text chunk
        :param begin: if it is the beginning of the text
        """
        if text_chunk is None or text_chunk == "":
            return
        if ctx.hidden:
            return

        pid = self.get_or_create_pid(meta)
        self.clear_chunks_input(pid)
        chunk = self.helpers.format_chunk(text_chunk)
        escaped_chunk = json.dumps(chunk)
        try:
            self.get_output_node(meta).page().runJavaScript(f"appendToInput({escaped_chunk});")
        except Exception as e:
            pass

    def append_node(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            html: str,
            type: int = 1,
            prev_ctx: Optional[CtxItem] = None,
            next_ctx: Optional[CtxItem] = None
    ):
        """
        Append and format raw text to output

        :param meta: context meta
        :param html: text to append
        :param type: type of message
        :param ctx: CtxItem instance
        :param prev_ctx: previous context item
        :param next_ctx: next context item
        """
        if ctx.hidden:
            return

        pid = self.get_or_create_pid(meta)
        self.append(
            pid,
            self.prepare_node(
                meta=meta,
                ctx=ctx,
                html=html,
                type=type,
                prev_ctx=prev_ctx,
                next_ctx=next_ctx
            )
        )

    def append(
            self,
            pid,
            html: str,
            flush: bool = False
    ):
        """
        Append text to output
        
        :param pid: ctx pid
        :param html: HTML code
        :param flush: True if flush only
        """
        if self.pids[pid].loaded and not self.pids[pid].use_buffer:
            self.clear_chunks(pid)
            self.flush_output(pid, html)  # render
            self.pids[pid].html = ""
        else:
            if not flush:
                self.pids[pid].html += html  # to buffer

    def append_context_item(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            prev_ctx: Optional[CtxItem] = None,
            next_ctx: Optional[CtxItem] = None
    ):
        """
        Append context item to output

        :param meta: context meta
        :param ctx: context item
        :param prev_ctx: previous context item
        :param next_ctx: next context item
        """
        self.append_input(
            meta,
            ctx,
            flush=False
        )
        self.append_output(
            meta,
            ctx,
            flush=False,
            prev_ctx=prev_ctx,
            next_ctx=next_ctx
        )  # + extra

    def append_extra(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            footer: bool = False,
            render: bool = True
    ) -> str:
        """
        Append extra data (images, files, etc.) to output
        
        :param meta: context meta
        :param ctx: context item
        :param footer: True if it is a footer
        :param render: True if render, False if only return HTML
        :return: HTML code
        """
        self.tool_output_end()  # reset tools

        pid = self.get_pid(meta)
        appended = []
        html = ""
        # images
        c = len(ctx.images)
        if c > 0:
            n = 1
            for image in ctx.images:
                if image is None:
                    continue
                # don't append if it is an external url
                # if image.startswith("http"):
                    # continue
                if image in appended or image in self.pids[pid].images_appended:
                    continue
                try:
                    appended.append(image)
                    html += self.body.get_image_html(image, n, c)
                    self.pids[pid].images_appended.append(image)
                    n += 1
                except Exception as e:
                    pass

        # files and attachments, TODO check attachments
        c = len(ctx.files)
        if c > 0:
            files_html = []
            n = 1
            for file in ctx.files:
                if file in appended or file in self.pids[pid].files_appended:
                    continue
                try:
                    appended.append(file)
                    files_html.append(self.body.get_file_html(file, n, c))
                    self.pids[pid].files_appended.append(file)
                    n += 1
                except Exception as e:
                    pass
            if files_html:
                html += "<br/>" + "<br/>".join(files_html)

        # urls
        c = len(ctx.urls)
        if c > 0:
            urls_html = []
            n = 1
            for url in ctx.urls:
                if url in appended or url in self.pids[pid].urls_appended:
                    continue
                try:
                    appended.append(url)
                    urls_html.append(self.body.get_url_html(url, n, c))
                    self.pids[pid].urls_appended.append(url)
                    n += 1
                except Exception as e:
                    pass
            if urls_html:
                html += "<br/>" + "<br/>".join(urls_html)

        # docs json
        if self.window.core.config.get('ctx.sources'):
            if ctx.doc_ids is not None and len(ctx.doc_ids) > 0:
                try:
                    docs = self.body.get_docs_html(ctx.doc_ids)
                    html += docs
                except Exception as e:
                    pass
        # flush
        if render and html != "":
            if footer:
                # append to output
                self.append(pid, html)
            else:
                # append to existing message box using JS
                escaped_html = json.dumps(html)
                self.get_output_node(meta).page().runJavaScript("appendExtra('{}',{});".format(ctx.id, escaped_html))

        return html

    def append_timestamp(
            self,
            ctx: CtxItem,
            text: str,
            type: Optional[int] = None
    ) -> str:
        """
        Append timestamp to text

        :param ctx: context item
        :param text: Input text
        :param type: Type of message
        :return: Text with timestamp (if enabled)
        """
        if ctx is not None and ctx.input_timestamp is not None:
            timestamp = None
            if type == self.NODE_INPUT:
                timestamp = ctx.input_timestamp
            elif type == self.NODE_OUTPUT:
                timestamp = ctx.output_timestamp
            if timestamp is not None:
                ts = datetime.fromtimestamp(timestamp)
                hour = ts.strftime("%H:%M:%S")
                text = '<span class="ts">{}: </span>{}'.format(hour, text)
        return text

    def reset(
            self,
            meta: Optional[CtxMeta] = None
    ):
        """
        Reset

        :param meta: Context meta
        """
        pid = self.get_pid(meta)
        if pid is not None and pid in self.pids:  # in PIDs only if at least one ctx item is appended
            self.reset_by_pid(pid)
        else:
            # there is no pid here if empty context so check for meta, and clear current
            if meta is not None:
                # create new PID using only meta
                pid = self.get_or_create_pid(meta)
                self.reset_by_pid(pid)

    def reset_by_pid(self, pid: Optional[int]):
        """
        Reset by PID

        :param pid: context PID
        """
        self.parser.reset()
        self.pids[pid].item = None
        self.pids[pid].html = ""
        self.clear_nodes(pid)
        self.clear_chunks(pid)
        self.pids[pid].images_appended = []
        self.pids[pid].urls_appended = []
        self.pids[pid].files_appended = []
        self.get_output_node_by_pid(pid).reset_current_content()
        self.reset_names_by_pid(pid)

    def clear_input(self):
        """Clear input"""
        self.get_input_node().clear()

    def clear_output(
            self,
            meta: Optional[CtxMeta] = None
    ):
        """
        Clear output

        :param meta: Context meta
        """
        self.reset(meta)

    def clear_chunks(self, pid):
        """
        Clear chunks from output

        :param pid: context PID
        """
        if pid is None:
            return
        self.clear_chunks_input(pid)
        self.clear_chunks_output(pid)

    def clear_chunks_input(
            self,
            pid: Optional[int]
    ):
        """
        Clear chunks from input

        :pid: context PID
        """
        if pid is None:
            return
        if not self.pids[pid].loaded:
            js = "var element = document.getElementById('_append_input_');"
            js += "if (element) { element.innerHTML = ''; }"
        else:
            js = "clearInput();"
        try:
            self.get_output_node_by_pid(pid).page().runJavaScript(js)
        except Exception as e:
            pass

    def clear_chunks_output(
            self,
            pid: Optional[int]
    ):
        """
        Clear chunks from output

        :pid: context PID
        """
        if not self.pids[pid].loaded:
            js = "var element = document.getElementById('_append_output_');"
            js += "if (element) { element.innerHTML = ''; }"
        else:
            js = "clearOutput();"
        try:
            self.get_output_node_by_pid(pid).page().runJavaScript(js)
        except Exception as e:
            pass

    def clear_nodes(
            self,
            pid: Optional[int]
    ):
        """
        Clear nodes from output

        :pid: context PID
        """
        if not self.pids[pid].loaded:
            js = "var element = document.getElementById('_nodes_');"
            js += "if (element) { element.innerHTML = ''; }"
        else:
            js = "clearNodes();"
        try:
            self.get_output_node_by_pid(pid).page().runJavaScript(js)
        except Exception as e:
            pass

    def prepare_node(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            html: str,
            type: int = 1,
            prev_ctx: Optional[CtxItem] = None,
            next_ctx: Optional[CtxItem] = None
    ) -> str:
        """
        Prepare node HTML
        
        :param meta: context meta
        :param ctx: CtxItem instance
        :param html: html text
        :param type: type of message
        :param prev_ctx: previous context item
        :param next_ctx: next context item
        :return: prepared HTML
        """
        pid = self.get_or_create_pid(meta)
        if type == self.NODE_OUTPUT:
            return self.prepare_node_output(
                meta=meta,
                ctx=ctx,
                html=html,
                prev_ctx=prev_ctx,
                next_ctx=next_ctx
            )
        elif type == self.NODE_INPUT:
            return self.prepare_node_input(
                pid=pid,
                ctx=ctx,
                html=html,
                prev_ctx=prev_ctx,
                next_ctx=next_ctx
            )

    def prepare_node_input(
            self,
            pid,
            ctx: CtxItem,
            html: str,
            prev_ctx: Optional[CtxItem] = None,
            next_ctx: Optional[CtxItem] = None
    ) -> str:
        """
        Prepare input node

        :param pid: context PID
        :param ctx: CtxItem instance
        :param html: html text
        :param prev_ctx: previous context item
        :param next_ctx: next context item
        :return: prepared HTML
        """
        msg_id = "msg-user-" + str(ctx.id) if ctx is not None else ""
        content = self.append_timestamp(
            ctx,
            self.helpers.format_user_text(html),
            type=self.NODE_INPUT
        )
        html = "<p>" + content + "</p>"
        html = self.helpers.post_format_text(html)
        name = self.pids[pid].name_user

        if ctx.internal and ctx.input.startswith("[{"):
            name = trans("msg.name.system")
        if type(ctx.extra) is dict and "agent_evaluate" in ctx.extra:
            name = trans("msg.name.evaluation")

        # debug
        debug = ""
        if self.is_debug():
            debug = self.append_debug(ctx, pid, "input")

        extra = ""
        if ctx.extra is not None and "footer" in ctx.extra:
            extra = ctx.extra["footer"]
        html = (
            '<div class="msg-box msg-user" id="{msg_id}">'
            '<div class="name-header name-user">{name}</div>'
            '<div class="msg">'
            '{html}'
            '<div class="msg-extra">{extra}</div>'
            '{debug}'
            '</div>'
            '</div>'
        ).format(
            msg_id=msg_id,
            name=name,
            html=html,
            extra=extra,
            debug=debug,
        )

        return html

    def prepare_node_output(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            html: str,
            prev_ctx: Optional[CtxItem] = None,
            next_ctx: Optional[CtxItem] = None
    ) -> str:
        """
        Prepare output node

        :param meta: context meta
        :param ctx: CtxItem instance
        :param html: html text
        :param prev_ctx: previous context item
        :param next_ctx: next context item
        :return: prepared HTML
        """
        is_cmd = False
        if (
                next_ctx is not None and
                next_ctx.internal and
                (len(ctx.cmds) > 0 or (ctx.extra_ctx is not None and len(ctx.extra_ctx) > 0))
        ):
            is_cmd = True
        pid = self.get_or_create_pid(meta)
        msg_id = "msg-bot-" + str(ctx.id) if ctx is not None else ""
        # if is_cmd:
        # html = self.helpers.format_cmd_text(html)
        html = self.helpers.pre_format_text(html)
        html = self.append_timestamp(ctx, html, type=self.NODE_OUTPUT)
        html = self.parser.parse(html)
        html = self.helpers.post_format_text(html)
        extra = self.append_extra(meta, ctx, footer=True, render=False)
        footer = self.body.prepare_action_icons(ctx)

        # append tool output
        tool_output = ""
        spinner = ""
        icon = os.path.join(
            self.window.core.config.get_app_path(),
            "data", "icons", "expand.svg"
        )
        output_class = "display:none"
        cmd_icon = (
            '<img src="file://{}" width="25" height="25" valign="middle">'
            .format(icon)
        )
        expand_btn = (
            "<span class='toggle-cmd-output' onclick='toggleToolOutput({});' "
            "role='button'>{} {}</span>"
            .format(str(ctx.id), cmd_icon, trans('action.cmd.expand'))
        )

        # check if next ctx is internal and current ctx has commands
        if is_cmd:
            # first, check current input if agent step and results
            if ctx.results is not None and len(ctx.results) > 0 \
                    and isinstance(ctx.extra, dict) and "agent_step" in ctx.extra:
                tool_output = self.helpers.format_cmd_text(str(ctx.input))
                output_class = ""  # show tool output
            else:
                # get output from next input (JSON response)
                tool_output = self.helpers.format_cmd_text(str(next_ctx.input))
                output_class = ""  # show tool output

        # check if agent step and results in current ctx
        elif ctx.results is not None and len(ctx.results) > 0 \
                and isinstance(ctx.extra, dict) and "agent_step" in ctx.extra:
            tool_output = self.helpers.format_cmd_text(str(ctx.input))
        else:
            # loading spinner
            if (
                    next_ctx is None and
                    (
                            ctx.output.startswith("~###~{\"cmd\"") or
                            ctx.output.strip().endswith("}~###~") or
                            len(ctx.cmds) > 0
                    )
            ):
                spinner_class = "display:none"  # hide by default
                if ctx.live:
                    spinner_class = ""  # show spinner only if commands and active run
                icon = os.path.join(
                    self.window.core.config.get_app_path(),
                    "data", "icons", "sync.svg"
                )
                spinner = (
                    '<span class="spinner" style="{}">'
                    '<img src="file://{}" width="30" height="30" '
                    'class="loading"></span>'
                    .format(spinner_class, icon)
                )

        html_tools = (
                spinner +
                '<div class="tool-output" style="{}">'.format(output_class) +
                expand_btn +
                '<div class="content" style="display:none">' +
                tool_output +
                '</div></div>'
        )
        tool_extra = self.body.prepare_tool_extra(ctx)

        # debug
        debug = ""
        if self.is_debug():
            debug = self.append_debug(ctx, pid, "output")

        html = (
            '<div class="msg-box msg-bot" id="{msg_id}">'
            '<div class="name-header name-bot">{name_bot}</div>'
            '<div class="msg">'
            '{html}'
            '<div class="msg-tool-extra">{tool_extra}</div>'
            '{html_tools}'
            '<div class="msg-extra">{extra}</div>'
            '{footer}'
            '{debug}'
            '</div>'
            '</div>'
        ).format(
            msg_id=msg_id,
            name_bot=self.pids[pid].name_bot,
            html=html,
            html_tools=html_tools,
            extra=extra,
            footer=footer,
            tool_extra=tool_extra,
            debug=debug,
        )

        return html

    def flush_output(
            self,
            pid: Optional[int],
            html: str
    ):
        """
        Flush output

        :param pid: context PID
        :param html: HTML code
        """
        escaped_html = json.dumps(html)
        try:
            node = self.get_output_node_by_pid(pid)
            node.page().runJavaScript(f"appendNode({escaped_html});")
            node.update_current_content()
        except Exception as e:
            pass

    def reload(self):
        """Reload output, called externally only on theme change to redraw content"""
        self.window.controller.ctx.refresh_output()  # if clear all and appends all items again

    def flush(
            self,
            pid:
            Optional[int]
    ):
        """
        Flush output

        :param pid: context PID
        """
        if self.pids[pid].loaded:
            return  # wait for page load

        html = self.body.get_html(pid)
        self.pids[pid].document = html
        self.get_output_node_by_pid(pid).setHtml(html, baseUrl="file://")

    def get_output_node(
            self,
            meta: Optional[CtxMeta] = None
    ):
        """
        Get output node

        :param meta: context meta
        :return: output node
        """
        return self.window.core.ctx.output.get_current(meta)

    def get_output_node_by_pid(
            self,
            pid: Optional[int]
    ):
        """
        Get output node by PID

        :param pid: context pid
        :return: output node
        """
        return self.window.core.ctx.output.get_by_pid(pid)

    def get_input_node(self) -> ChatInput:
        """
        Get input node

        :return: input node
        """
        return self.window.ui.nodes['input']

    def get_document(
            self,
            plain: bool = False
    ):
        """
        Get document content (plain or HTML)

        :param plain: True to convert to plain text
        :return: document content
        """
        pid = self.window.core.ctx.container.get_active_pid()
        if pid is None:
            return ""
        if plain:
            return self.parser.to_plain_text(self.pids[pid].document.replace("<br>", "\n"))
        return self.pids[pid].document

    def remove_item(self, ctx: CtxItem):
        """
        Remove item from output

        :param ctx: context item
        """
        try:
            self.get_output_node(ctx.meta).page().runJavaScript("removeNode({});".format(ctx.id))
        except Exception as e:
            pass

    def remove_items_from(self, ctx: CtxItem):
        """
        Remove item from output

        :param ctx: context item
        """
        try:
            self.get_output_node(ctx.meta).page().runJavaScript("removeNodesFromId({});".format(ctx.id))
        except Exception as e:
            pass

    def reset_names(self, meta: CtxMeta):
        """
        Reset names

       :param meta: context meta
        """
        if meta is None:
            return
        pid = self.get_or_create_pid(meta)
        self.reset_names_by_pid(pid)

    def reset_names_by_pid(self, pid: int):
        """
        Reset names by PID

        :param pid: context pid
        """
        self.pids[pid].name_user = trans("chat.name.user")
        self.pids[pid].name_bot = trans("chat.name.bot")

    def on_reply_submit(self, ctx: CtxItem):
        """
        On regenerate submit

        :param ctx: context item
        """
        # remove all items from ID
        self.remove_items_from(ctx)

    def on_edit_submit(self, ctx: CtxItem):
        """
        On regenerate submit

        :param ctx: context item
        """
        # remove all items from ID
        self.remove_items_from(ctx)

    def on_enable_edit(self, live: bool = True):
        """
        On enable edit icons

        :param live: True if live update
        """
        if not live:
            return
        try:
            nodes = self.get_all_nodes()
            for node in nodes:
                node.page().runJavaScript("enableEditIcons();")
        except Exception as e:
            pass

    def on_disable_edit(self, live: bool = True):
        """
        On disable edit icons

        :param live: True if live update
        """
        if not live:
            return
        try:
            nodes = self.get_all_nodes()
            for node in nodes:
                node.page().runJavaScript("disableEditIcons();")
        except Exception as e:
            pass

    def on_enable_timestamp(self, live: bool = True):
        """
        On enable timestamp

        :param live: True if live update
        """
        if not live:
            return
        try:
            nodes = self.get_all_nodes()
            for node in nodes:
                node.page().runJavaScript("enableTimestamp();")
        except Exception as e:
            pass

    def on_disable_timestamp(self, live: bool = True):
        """
        On disable timestamp

        :param live: True if live update
        """
        if not live:
            return
        try:
            nodes = self.get_all_nodes()
            for node in nodes:
                node.page().runJavaScript("disableTimestamp();")
        except Exception as e:
            pass

    def update_names(
            self,
            meta: CtxMeta,
            ctx: CtxItem
    ):
        """
        Update names
        
        :param meta: context meta
        :param ctx: context item
        """
        pid = self.get_or_create_pid(meta)
        if ctx.input_name is not None and ctx.input_name != "":
            self.pids[pid].name_user = ctx.input_name
        if ctx.output_name is not None and ctx.output_name != "":
            self.pids[pid].name_bot = ctx.output_name

    def clear_all(self):
        """Clear all"""
        for pid in self.pids:
            self.clear_chunks(pid)
            self.clear_nodes(pid)
            self.pids[pid].html = ""

    def scroll_to_bottom(self):
        """Scroll to bottom"""
        pass

    def append_block(self):
        """Append block to output"""
        pass

    def to_end(self, ctx: CtxItem):
        """
        Move cursor to end of output

        :param ctx: context item
        """
        pass

    def get_all_nodes(self) -> list:
        """
        Return all registered nodes

        :return: list of ChatOutput nodes (tabs)
        """
        return self.window.core.ctx.output.get_all()

    # TODO: on lang change

    def reload_css(self):
        """Reload CSS - all, global"""
        to_json = json.dumps(self.body.prepare_styles())
        nodes = self.get_all_nodes()
        for pid in self.pids:
            if self.pids[pid].loaded:
                for node in nodes:
                    node.page().runJavaScript("updateCSS({});".format(to_json))
                    if self.window.core.config.get('render.blocks'):
                        node.page().runJavaScript("enableBlocks();")
                    else:
                        node.page().runJavaScript("disableBlocks();")  # TODO: ctx!!!!!
                return

    def on_theme_change(self):
        """On theme change"""
        self.window.controller.theme.markdown.load()
        for pid in self.pids:
            if self.pids[pid].loaded:
                self.reload_css()
                return

    def tool_output_append(
            self,
            meta: CtxMeta,
            content: str
    ):
        """
        Add tool output (append)

        :param meta: context meta
        :param content: content
        """
        escaped_content = json.dumps(content)
        try:
            self.get_output_node(meta).page().runJavaScript(f"appendToolOutput({escaped_content});")
        except Exception as e:
            pass

    def tool_output_update(
            self,
            meta: CtxMeta,
            content: str
    ):
        """
        Replace tool output

        :param meta: context meta
        :param content: content
        """
        escaped_content = json.dumps(content)
        try:
            self.get_output_node(meta).page().runJavaScript(f"updateToolOutput({escaped_content});")
        except Exception as e:
            pass

    def tool_output_clear(self, meta: CtxMeta):
        """
        Clear tool output

        :param meta: context meta
        """
        try:
            self.get_output_node(meta).page().runJavaScript(f"clearToolOutput();")
        except Exception as e:
            pass

    def tool_output_begin(self, meta: CtxMeta):
        """
        Begin tool output

        :param meta: context meta
        """
        try:
            self.get_output_node(meta).page().runJavaScript(f"beginToolOutput();")
        except Exception as e:
            pass

    def tool_output_end(self):
        """End tool output"""
        try:
            self.get_output_node().page().runJavaScript(f"endToolOutput();")
        except Exception as e:
            pass

    def append_debug(
            self,
            ctx: CtxItem,
            pid,
            title: Optional[str] = None
    ) -> str:
        """
        Append debug info

        :param ctx: context item
        :param pid: context PID
        :param title: debug title
        """
        if title is None:
            title = "debug"
        debug = "<b>" +title+ ":</b> pid: "+str(pid)+", ctx: " + str(ctx.to_dict())
        return "<div class='debug'>" + debug + "</div>"

    def is_debug(self) -> bool:
        """
        Check if debug mode is enabled

        :return: True if debug mode is enabled
        """
        return self.window.core.config.get("debug.render", False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.24 22:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QCheckBox, QHBoxLayout, QWidget

from pygpt_net.ui.widget.anims.toggles import AnimToggle
from pygpt_net.utils import trans


class OptionCheckbox(QWidget):
    def __init__(self, window=None, parent_id: str = None, id: str = None, option: dict = None):
        """
        Settings checkbox

        :param window: main window
        :param id: option id
        :param parent_id: parent option id
        :param option: option data
        """
        super(OptionCheckbox, self).__init__(window)
        self.window = window
        self.id = id
        self.parent_id = parent_id
        self.option = option
        self.value = False
        self.title = ""
        self.real_time = False

        # init from option data
        if self.option is not None:
            if "label" in self.option and self.option["label"] is not None \
                    and self.option["label"] != "":
                self.title = trans(self.option["label"])
            if "value" in self.option:
                self.value = self.option["value"]
            if "real_time" in self.option:
                self.real_time = self.option["real_time"]

        self.box = AnimToggle(self.title, self.window)
        if self.value is not None:
            self.box.setChecked(self.value)

        self.box.stateChanged.connect(
            lambda: self.window.controller.config.checkbox.on_update(
                self.parent_id,
                self.id,
                self.option,
                self.box.isChecked()
            )
        )

        self.layout = QHBoxLayout()
        self.layout.addWidget(self.box)

        self.setLayout(self.layout)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Any

from pygpt_net.core.events import KernelEvent
from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans


class Llama:
    def __init__(self, window=None):
        """
        Agents (Llama-index) controller

        :param window: Window instance
        """
        self.window = window
        self.eval_step = 0
        self.options = {
            "agent.llama.loop.score": {
                "type": "int",
                "slider": True,
                "label": "agent.llama.loop.score",
                "min": 0,
                "max": 100,
                "step": 1,
                "value": 75,
                "multiplier": 1,
            },
        }

    def setup(self):
        """Setup agent controller"""
        # register hooks
        self.window.ui.add_hook("update.global.agent.llama.loop.score", self.hook_update)
        self.reload()  # restore config

    def reload(self):
        """Reload agent toolbox options"""
        # loop enable checkbox
        if self.window.core.config.get('agent.llama.loop.enabled'):
            self.window.ui.config['global']['agent.llama.loop.enabled'].setChecked(True)
        else:
            self.window.ui.config['global']['agent.llama.loop.enabled'].setChecked(False)

        # loop score slider
        self.window.controller.config.apply_value(
            parent_id="global",
            key="agent.llama.loop.score",
            option=self.options["agent.llama.loop.score"],
            value=self.window.core.config.get('agent.llama.loop.score'),
        )

    def reset_eval_step(self):
        """Reset evaluation step"""
        self.eval_step = 0

    def eval_step_next(self):
        """Next evaluation step"""
        self.eval_step += 1

    def get_eval_step(self) -> int:
        """
        Get evaluation step

        :return: evaluation step
        """
        return self.eval_step

    def on_user_send(self, text: str):
        """
        Run begin

        :param text: user input
        """
        self.reset_eval_step()  # reset evaluation step

    def on_end(self):
        """End of run"""
        self.eval_step = 0  # reset evaluation step
        if self.window.core.config.get("agent.goal.notify"):
            self.window.ui.tray.show_msg(
                trans("notify.agent.goal.title"),
                trans("notify.agent.goal.content"),
            )

    def on_finish(self, ctx: CtxItem):
        """
        Finish agent run

        :param ctx: CtxItem
        """
        if not self.window.core.config.get("agent.llama.loop.enabled"):
            self.on_end()
            return  # abort if loop is disabled

        # check if not stopped
        if self.window.controller.kernel.stopped():
            self.on_end()
            return

        # check max steps
        max_steps = int(self.window.core.config.get("agent.llama.max_eval"))
        if max_steps != 0 and self.get_eval_step() >= max_steps:
            self.on_end()
            return  # abort if max steps reached

        # evaluation step++
        self.eval_step_next()

        context = BridgeContext()
        context.ctx = ctx
        context.history = self.window.core.ctx.all(meta_id=ctx.meta.id)
        self.window.update_status(trans('status.evaluating'))  # show info
        event = KernelEvent(KernelEvent.REQUEST_NEXT, {
            'context': context,
            'extra': {},
        })
        self.window.dispatch(event)

    def on_stop(self):
        """Stop agent"""
        pass

    def update(self):
        """Update agent status"""
        pass

    def hook_update(self, key: str, value: Any, caller, *args, **kwargs):
        """
        Hook: on option update

        :param key: config key
        :param value: config value
        :param caller: caller name
        :param args: args
        :param kwargs: kwargs
        """
        if self.window.core.config.get(key) == value:
            return
        if key == 'agent.llama.loop.score':
            self.window.core.config.set(key, int(value))
            self.window.core.config.save()
            self.update()